# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: Apache-2.0

import pathlib
import sys


def argparse_create():
    import argparse

    parser = argparse.ArgumentParser(
        prog=pathlib.Path(sys.argv[0]).name + " --command maketx",
        description=(
            "Generate Cycles texture cache (.tx) files.\n"
            "\n"
            "When a file path is given, generate a tx file for that image.\n"
            "When no file path is given, a blend file must be loaded and tx\n"
            "files are generated for all images used by shader node trees."
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "filepath",
        type=str,
        nargs="?",
        default=None,
        help="Source image file path. If omitted, generate for all images in the blend file.",
    )

    parser.add_argument(
        "--colorspace",
        dest="colorspace",
        type=str,
        default="auto",
        help="Color space name. Auto-detected if omitted. Only used with a file path.",
    )

    parser.add_argument(
        "--alpha-type",
        dest="alpha_type",
        type=str,
        default="auto",
        choices=["straight", "premultiplied", "channel_packed", "none", "auto"],
        help="Alpha type. Auto-detected if omitted. Only used with a file path.",
    )

    parser.add_argument(
        "--cache-dir",
        dest="cache_dir",
        type=str,
        default="",
        help="Output directory for tx files, relative to source image or absolute.",
    )

    return parser


def maketx_command(argv):
    parser = argparse_create()
    args = parser.parse_args(argv)

    if args.filepath is not None:
        return maketx_file(args)
    else:
        return maketx_blend(args, parser)


def maketx_file(args):
    import _cycles

    filepath = str(pathlib.Path(args.filepath).absolute())

    try:
        out_filepath = _cycles.maketx(
            filepath,
            colorspace=args.colorspace,
            alpha_type=args.alpha_type,
            cache_dir=args.cache_dir,
        )
    except RuntimeError as ex:
        sys.stderr.write("Error: {:s}\n".format(str(ex)))
        return 1

    print(out_filepath)
    return 0


def maketx_blend(args, parser):
    import bpy

    if not bpy.data.filepath:
        parser.error("No image file path given and no blend file loaded.")

    if args.colorspace != "auto":
        parser.error("--colorspace is only used with an image file path argument.")

    if args.alpha_type != "auto":
        parser.error("--alpha-type is only used with an image file path argument.")

    if args.cache_dir:
        bpy.context.preferences.filepaths.texture_cache_directory = args.cache_dir

    bpy.ops.render.generate_texture_cache()
    return 0



# SPDX-FileCopyrightText: 2018-2021 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

def export_user_extensions(hook_name, export_settings, *args):
    if args and hasattr(args[0], "extensions"):
        if args[0].extensions is None:
            args[0].extensions = {}

    for extension in export_settings['gltf_user_extensions']:
        hook = getattr(extension, hook_name, None)
        if hook is not None:
            try:
                hook(*args, export_settings)
            except Exception as e:
                if getattr(extension, 'is_critical', False):
                    export_settings['log'].error(
                        "Critical extension hook " +
                        hook_name +
                        " fails on " +
                        extension.__module__ +
                        ": " +
                        str(e),
                        popup=True)
                    raise RuntimeError("Export aborted due to critical extension failure") from e
                else:
                    export_settings['log'].error("Extension hook " + hook_name + " fails on " + extension.__module__)
                    export_settings['log'].error(str(e))



# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import io
from pathlib import Path
from typing import Callable

__all__ = (
    'mutex_lock',
    'mutex_unlock',
)


# Dictionary of local library path to a tuple with:
# - lock file handle
# - path of the lock file
# - unlock function
_mutex_locks: dict[Path, tuple[io.IOBase, Path, Callable[[io.IOBase], None]]] = {}

_registered_atexit = False


def mutex_lock(local_library_path: Path) -> bool:
    """Lock the library for syncing.

    Create a file on disk that signals to other Blender instances that this
    remote asset library is being synced by this Blender.

    This uses approaches from:
        - https://www.pythontutorials.net/blog/make-sure-only-a-single-instance-of-a-program-is-running/
        - https://yakking.branchable.com/posts/procrun-2-pidfiles/

    :returns: true if the lock was created successfully, false if some other
        Blender already locked this library.
    """
    global _registered_atexit

    import atexit
    import sys

    if not _registered_atexit:
        atexit.register(_unlock_all)
        _registered_atexit = True

    # Choose platform-dependent _obtain_lock(file) and _release_lock() functions.
    if sys.platform == "win32":
        import msvcrt

        def _obtain_lock(file: io.IOBase) -> None:
            # Lock the first byte of the file (arbitrary choice)
            msvcrt.locking(file.fileno(), msvcrt.LK_NBLCK, 1)

        def _release_lock(file: io.IOBase) -> None:
            msvcrt.locking(file.fileno(), msvcrt.LK_UNLCK, 1)
    else:
        import fcntl

        def _obtain_lock(file: io.IOBase) -> None:
            fcntl.flock(file, fcntl.LOCK_EX | fcntl.LOCK_NB)

        def _release_lock(file: io.IOBase) -> None:
            # Closing the file automatically releases the lock.
            pass

    assert isinstance(local_library_path, Path)
    assert local_library_path not in _mutex_locks, "Locks are not reentrant"

    lockfile_path = local_library_path / "_sync.lock"

    # It is not suitable here to use an 'exclusive create' ('x' option) here.
    # That will still create a race condition, with the space between creation
    # of the file and locking it. So, better to make the existence of the file
    # meaningless, and only communicate the lock state with an actual file-system
    # lock.
    try:
        # Binary mode (`wb`) is required on Windows, for the locking.
        lockfile = lockfile_path.open('wb')
    except OSError:
        # on Windows, opening a file for writing, while another process already has it open, can fail.
        # That just means somebody else has ownership of it.
        return False
    try:
        _obtain_lock(lockfile)
    except OSError:
        # Lock is already held by another Blender.
        lockfile.close()
        return False

    # We have obtained an exclusive lock, which the OS will release when this
    # process is killed.
    _mutex_locks[local_library_path] = (lockfile, lockfile_path, _release_lock)
    return True


def mutex_unlock(local_library_path: Path) -> None:
    """Remove the lock created by mutex_lock(local_library_path)."""

    assert isinstance(local_library_path, Path)
    assert local_library_path in _mutex_locks, "library was not locked"

    lockfile, lockfile_path, release_lock = _mutex_locks[local_library_path]
    release_lock(lockfile)
    lockfile.close()

    del _mutex_locks[local_library_path]

    try:
        lockfile_path.unlink(missing_ok=True)
    except IOError:
        # Ignore errors when deleting the file. By now another process may have
        # recreated it and locked it again.
        pass


def _unlock_all() -> None:
    """Unlock all file mutexes.

    This is automatically called when the Python interpreter exits.

    From the OS perspective it's not necessary, as all locks are automatically
    released when the process stops. However, Python will complain with a
    ResourceWarning if any open files are not closed.
    """

    for local_library_path in list(_mutex_locks.keys()):
        mutex_unlock(local_library_path)



# SPDX-FileCopyrightText: 2009-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# Originally written by Matt Ebb

import bpy
from bpy.types import Operator
from bpy.app.translations import pgettext_rpt as rpt_


def guess_player_path(preset):
    import sys

    found_version = (0, 0, 0)

    if preset == 'INTERNAL':
        return bpy.app.binary_path, found_version

    elif preset == 'DJV':
        player_path = "djv"
        if sys.platform == "linux":
            found_version = (3, 4, 0)  # Assume it is at least 3.4.0
        elif sys.platform == "darwin":
            import os
            djv3_path = "/Applications/DJV.app/Contents/MacOS/DJV"
            djv2_path = "/Applications/DJV2.app/Contents/Resources/bin/djv"
            if os.path.exists(djv3_path):
                player_path = djv3_path
                found_version = (3, 4, 0)  # Assume it is at least 3.4.0
            elif os.path.exists(djv2_path):
                player_path = djv2_path
                found_version = (2, 0, 0)
        elif sys.platform == "win32":
            import winreg

            reg_value = None
            try:
                def extract_version(key_name):
                    """Given a key_name like "key name 3.3.4" extract version as (3, 3, 4)"""
                    version = None
                    parts = key_name.rsplit(" ", 1)
                    if len(parts) == 2:
                        try:
                            version = tuple(int(x) for x in parts[1].split("."))
                        except ValueError:
                            pass
                    return version

                # Enumerate versioned subkeys (e.g. "DJV 3.3.4", "DJV 3.4.0", etc.) and
                # pick the key with the greatest version.
                reg_base = r"SOFTWARE\WOW6432Node\Grizzly Peak 3D"
                with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, reg_base, 0, winreg.KEY_READ) as base_key:
                    best_version = None
                    best_subkey = None
                    index = 0
                    while True:
                        try:
                            subkey_name = winreg.EnumKey(base_key, index)
                        except OSError:
                            break
                        index += 1

                        if (version := extract_version(subkey_name)) is not None:
                            if best_version is None or version > best_version:
                                best_subkey = subkey_name
                                best_version = version
                                found_version = version

                    if best_subkey is not None:
                        reg_path = reg_base + "\\" + best_subkey
                        with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, reg_path, 0, winreg.KEY_READ) as regkey:
                            reg_value = winreg.QueryValue(regkey, None)
            except OSError:
                pass

            # Fallback to djv2 if we didn't find anything
            if not reg_value:
                try:
                    reg_path = r"SOFTWARE\Classes\djv\shell\open\command"
                    with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, reg_path, 0, winreg.KEY_READ) as regkey:
                        reg_value = winreg.QueryValue(regkey, None)
                        found_version = (2, 0, 0)
                except OSError:
                    pass

            if reg_value:
                if found_version > (2, 0, 0):
                    player_path = reg_value.strip() + "\\bin\\djv.exe"
                else:
                    binary = "djv.exe"
                    index = reg_value.find(binary)
                    if index > 0:
                        player_path = reg_value[:index + len(binary)]

    elif preset == 'FRAMECYCLER':
        player_path = "framecycler"

    elif preset == 'RV':
        player_path = "rv"

    elif preset == 'MPLAYER':
        player_path = "mplayer"

    else:
        player_path = ""

    return player_path, found_version


class PlayRenderedAnim(Operator):
    """Play back rendered frames/movies using an external player"""
    bl_idname = "render.play_rendered_anim"
    bl_label = "Play Rendered Animation"
    bl_options = {'REGISTER'}

    @staticmethod
    def _frame_path_with_number_char(rd, ch, **kwargs):
        # Replace the number with `ch`.

        # NOTE: make an api call for this would be nice, however this isn't needed in many places.
        file_a = rd.frame_path(frame=0, **kwargs)
        file_b = rd.frame_path(frame=-1, **kwargs)
        assert len(file_b) == len(file_a) + 1

        for number_beg in range(len(file_a)):
            if file_a[number_beg] != file_b[number_beg]:
                break

        for number_end in range(-1, -(len(file_a) + 1), -1):
            if file_a[number_end] != file_b[number_end]:
                break

        number_end += len(file_a) + 1
        return file_a[:number_beg] + (ch * (number_end - number_beg)) + file_a[number_end:]

    def execute(self, context):
        import os
        import subprocess
        from shlex import quote

        scene = context.scene
        rd = scene.render
        prefs = context.preferences
        fps_final = rd.fps / rd.fps_base

        preset = prefs.filepaths.animation_player_preset
        # file_path = bpy.path.abspath(rd.filepath)  # UNUSED
        is_movie = rd.is_movie_format

        views_format = rd.image_settings.views_format
        if rd.use_multiview and views_format == 'INDIVIDUAL':
            view_suffix = rd.views.active.file_suffix
        else:
            view_suffix = ""

        found_version = (0, 0, 0)

        # try and guess a command line if it doesn't exist
        if preset == 'CUSTOM':
            player_path = prefs.filepaths.animation_player
        else:
            player_path, found_version = guess_player_path(preset)

        if is_movie is False and preset in {'FRAMECYCLER', 'RV', 'MPLAYER'}:
            file = PlayRenderedAnim._frame_path_with_number_char(rd, "#", view=view_suffix)
            file = bpy.path.abspath(file)  # expand '//'
        else:
            path_valid = True
            # works for movies and images
            file = rd.frame_path(frame=scene.frame_start, preview=scene.use_preview_range, view=view_suffix)
            file = bpy.path.abspath(file)  # expand '//'
            if not os.path.exists(file):
                err_msg = rpt_("File {!r} not found").format(file)
                self.report({'WARNING'}, err_msg)
                path_valid = False

            # one last try for full range if we used preview range
            if scene.use_preview_range and not path_valid:
                file = rd.frame_path(frame=scene.frame_start, preview=False, view=view_suffix)
                file = bpy.path.abspath(file)  # expand '//'
                err_msg = rpt_("File {!r} not found").format(file)
                if not os.path.exists(file):
                    self.report({'WARNING'}, err_msg)

        remove_OCIO_env = False
        cmd = [player_path]
        # extra options, fps controls etc.
        if scene.use_preview_range:
            frame_start = scene.frame_preview_start
            frame_end = scene.frame_preview_end
        else:
            frame_start = scene.frame_start
            frame_end = scene.frame_end
        if preset == 'INTERNAL':
            # Use the current GPU backend for the player.
            import gpu
            gpu_backend = gpu.platform.backend_type_get()
            if gpu_backend not in {'NONE', 'UNKNOWN'}:
                cmd.extend([
                    "--gpu-backend", gpu_backend.lower(),
                ])
            del gpu, gpu_backend

            opts = [
                "-a",
                "-f", str(rd.fps), str(rd.fps_base),
                "-s", str(frame_start),
                "-e", str(frame_end),
                "-j", str(scene.frame_step),
                "-c", str(prefs.system.memory_cache_limit),
                file,
            ]
            cmd.extend(opts)
        elif preset == 'DJV':
            if found_version >= (3, 4, 0):
                opts = [
                    file,
                    "-speed", str(fps_final),
                    "-in", str(frame_start),
                    "-out", str(frame_end),
                    "-seek", str(scene.frame_current),
                    "-timeUnits", "Frames",
                ]
            elif found_version >= (3, 0, 0):
                opts = [
                    file,
                    "-speed", str(fps_final),
                ]
            else:
                remove_OCIO_env = True
                opts = [
                    file,
                    "-speed", str(fps_final),
                    "-in_out", str(frame_start), str(frame_end),
                    "-frame", str(scene.frame_current),
                    "-time_units", "Frames",
                ]
            cmd.extend(opts)
        elif preset == 'FRAMECYCLER':
            opts = [file, "{:d}-{:d}".format(scene.frame_start, scene.frame_end)]
            cmd.extend(opts)
        elif preset == 'RV':
            opts = ["-fps", str(rd.fps), "-play"]
            if scene.use_preview_range:
                opts += [
                    file.replace("#", "", file.count('#') - 1),
                    "{:d}-{:d}".format(frame_start, frame_end),
                ]
            else:
                opts.append(file)

            cmd.extend(opts)
        elif preset == 'MPLAYER':
            opts = []
            if is_movie:
                opts.append(file)
            else:
                opts += [
                    ("mf://" + file.replace("#", "?")),
                    "-mf",
                    "fps={:.4f}".format(fps_final),
                ]

            opts += ["-loop", "0", "-really-quiet", "-fs"]
            cmd.extend(opts)
        else:  # 'CUSTOM'
            cmd.append(file)

        # launch it
        print("Executing command:\n ", " ".join(quote(c) for c in cmd))

        try:
            env_copy = os.environ.copy()
            if remove_OCIO_env:
                env_copy.pop("OCIO", None)
            subprocess.Popen(cmd, env=env_copy)
        except Exception as ex:
            err_msg = rpt_("Couldn't run external animation player with command {!r}\n{:s}").format(cmd, str(ex))
            self.report(
                {'ERROR'},
                err_msg,
            )
            return {'CANCELLED'}

        return {'FINISHED'}


classes = (
    PlayRenderedAnim,
)



# SPDX-FileCopyrightText: 2018-2026 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0


import bpy
from .....io.com.gltf2_io_extensions import Extension
from .....io.com.constants import GLTF_IRIDESCENCE_IOR
from ...material import texture_info as gltf2_blender_gather_texture_info
from ..search_node_tree import \
    has_image_node_from_socket, \
    get_socket_from_gltf_material_node, \
    detect_iridescence_thickness_texure, \
    get_socket, \
    get_factor_from_socket


def export_iridescence(bmat, export_settings):

    # There is no iridescence Factor (Intensity) in Blender
    # So we consider that we export the extension if the Iridescence thickness is not 0

    # Mapping:
    # Iridescence Thickness (Blender) -> Iridescence Thickness Maximum (glTF)
    # Iridesence Thickness Mininum (Blender, glTF Output Material Node) -> Iridescence Thickness Minimum (glTF)
    # Iridescence Thickness Texture Red (Blender) -> Iridescence Thickness Texture (glTF)
    # Tickness IOR (Blender) -> Iridescence IOR (glTF) . Warning, Blender default 1.333, glTF default 1.3
    # IridescenceFactor (Blender, glTF Output Material Node) -> Iridescence Factor (glTF)
    # IridescenceTexture Red (Blender, glTF Output Material Node texture on
    # IridescenceFactor) -> Iridescence Texture (glTF)

    export_settings['current_texture_transform'] = {}
    iridescence_extension = {}
    uvmap_infos = {}
    udim_infos = {}

    # Cases where the extension is not exported:

    # No thickness socket found (no Principled Shader)
    iridescence_thickness_socket = get_socket(bmat.get_used_material().node_tree, "Thin Film Thickness")
    if iridescence_thickness_socket.socket is None:
        return None, {}, {}

    # factor (from glTF Output group node)
    iridescence_factor_socket = get_socket_from_gltf_material_node(
        bmat.get_used_material().node_tree, "Iridescence Factor")
    if iridescence_factor_socket.socket is None:
        return None, {}, {}

    # Thickness minimum (from glTF Output group node)
    iridescence_thickness_minimum_socket = get_socket_from_gltf_material_node(
        bmat.get_used_material().node_tree, "Iridescence Thickness Minimum")
    if iridescence_thickness_minimum_socket is None:
        return None, {}, {}

    # Thickness socket is not linked and default value is 0
    # But we have to export it anyway, because it can be animated and we need to export the path for KHR_animation_pointer
    # If not animated, it will be remove after export, because of the default value

    # Iridescence Factor socket is not linked and default value is 0
    # But we have to export it anyway, because it can be animated and we need to export the path for KHR_animation_pointer
    # If not animated, it will be remove after export, because of the default value

    # IOR
    iridescence_ior_socket = get_socket(bmat.get_used_material().node_tree, "Thin Film IOR")
    if abs(iridescence_ior_socket.socket.default_value - GLTF_IRIDESCENCE_IOR) > 0.0001:
        iridescence_extension['iridescenceIor'] = iridescence_ior_socket.socket.default_value

    # IOR storing for KHR_animation_pointer
    path_ = {}
    path_['length'] = 1
    path_['path'] = "/materials/XXX/extensions/KHR_materials_iridescence/iridescenceIor"
    export_settings['current_paths']["node_tree." +
                                     iridescence_ior_socket.socket.path_from_id() +
                                     ".default_value"] = path_

    # Iridescence Factor
    iridescence_factor, path = get_factor_from_socket(iridescence_factor_socket, kind='VALUE')
    if iridescence_factor != 0.0:
        iridescence_extension['iridescenceFactor'] = iridescence_factor

    # Iridescence Factor storing for KHR_animation_pointer
    path_ = {}
    path_['length'] = 1
    path_['path'] = "/materials/XXX/extensions/KHR_materials_iridescence/iridescenceFactor"
    export_settings['current_paths'][path] = path_

    # IridescenceTexture
    if has_image_node_from_socket(iridescence_factor_socket, export_settings) is True:

        texture_info, uvmap_info, udim_info, _ = gltf2_blender_gather_texture_info.gather_texture_info(
            iridescence_factor_socket,
            (iridescence_factor_socket,),
            export_settings
        )

        if texture_info is not None:
            iridescence_extension['iridescenceTexture'] = texture_info
            uvmap_infos.update({'iridescenceTexture': uvmap_info})
            udim_infos.update({'iridescenceTexture': udim_info} if len(udim_info) > 0 else {})

            if len(export_settings['current_texture_transform']) != 0:
                for k in export_settings['current_texture_transform'].keys():
                    path_ = {}
                    path_['length'] = export_settings['current_texture_transform'][k]['length']
                    path_['path'] = export_settings['current_texture_transform'][k]['path'].replace(
                        "YYY", "extensions/KHR_materials_iridescence/iridescenceTexture/extensions")
                    path_['vector_type'] = export_settings['current_texture_transform'][k]['vector_type']
                    if k in export_settings['current_paths']:
                        if 'additional' not in export_settings['current_paths'][k]:
                            export_settings['current_paths'][k]['additional'] = []
                        if path_['path'] != export_settings['current_paths'][k]['path']:
                            export_settings['current_paths'][k]['additional'].append(path_['path'])
                    else:
                        export_settings['current_paths'][k] = path_
            export_settings['current_texture_transform'] = {}

    # Iridescence Thickness Maximum
    thickness_non_linked = iridescence_thickness_socket.socket is not None and isinstance(
        iridescence_thickness_socket.socket, bpy.types.NodeSocket) and not iridescence_thickness_socket.socket.is_linked

    if thickness_non_linked is True:

        is_texture_iridescence = False

        if abs(iridescence_thickness_socket.socket.default_value - 400.0) > 0.0001:
            iridescence_extension['iridescenceThicknessMaximum'] = iridescence_thickness_socket.socket.default_value

        # Iridescence Thickness Maximum storing path for KHR_animation_pointer
        path_ = {}
        path_['length'] = 1
        path_['path'] = "/materials/XXX/extensions/KHR_materials_iridescence/iridescenceThicknessMaximum"
        export_settings['current_paths']["node_tree." + iridescence_thickness_socket.socket.path_from_id() +
                                         ".default_value"] = path_

    else:
        is_texture_iridescence, iridescence_data = detect_iridescence_thickness_texure(
            iridescence_thickness_socket, iridescence_thickness_minimum_socket, export_settings)
        if is_texture_iridescence:
            # Texture found, so export from data retrieved
            if abs(iridescence_data['thickness_maximum'].node.outputs[0].default_value - 400.0) > 0.0001:
                iridescence_extension['iridescenceThicknessMaximum'] = iridescence_data['thickness_maximum'].node.outputs[0].default_value

            # Iridescence Thickness Maximum storing path for KHR_animation_pointer
            path_ = {}
            path_['length'] = 1
            path_['path'] = "/materials/XXX/extensions/KHR_materials_iridescence/iridescenceThicknessMaximum"
            export_settings['current_paths']["node_tree." +
                                             iridescence_data['thickness_maximum'].path_from_id() +
                                             ".default_value"] = path_

            # Thickness Minimum
            if abs(iridescence_data['thickness_minimum'].node.outputs[0].default_value - 100.0) > 0.0001:
                iridescence_extension['iridescenceThicknessMinimum'] = iridescence_data['thickness_minimum'].node.outputs[0].default_value

            # Iridescence Thickness Minimum storing path for KHR_animation_pointer
            path_ = {}
            path_['length'] = 1
            path_['path'] = "/materials/XXX/extensions/KHR_materials_iridescence/iridescenceThicknessMinimum"
            export_settings['current_paths']["node_tree." +
                                             iridescence_data['thickness_minimum'].path_from_id() +
                                             ".default_value"] = path_

            # Texture
            if iridescence_data['tex_socket'] is not None:
                texture_info, uvmap_info, udim_info, _ = gltf2_blender_gather_texture_info.gather_texture_info(
                    iridescence_data['tex_socket'],
                    (iridescence_data['tex_socket'],),
                    export_settings,
                )

                if texture_info is not None:
                    iridescence_extension['iridescenceThicknessTexture'] = texture_info
                    uvmap_infos.update({'iridescenceThicknessTexture': uvmap_info})
                    udim_infos.update({'iridescenceThicknessTexture': udim_info} if len(udim_info) > 0 else {})

                    if len(export_settings['current_texture_transform']) != 0:
                        for k in export_settings['current_texture_transform'].keys():
                            path_ = {}
                            path_['length'] = export_settings['current_texture_transform'][k]['length']
                            path_['path'] = export_settings['current_texture_transform'][k]['path'].replace(
                                "YYY", "extensions/KHR_materials_iridescence/iridescenceThicknessTexture/extensions")
                            path_['vector_type'] = export_settings['current_texture_transform'][k]['vector_type']
                            if k in export_settings['current_paths']:
                                if 'additional' not in export_settings['current_paths'][k]:
                                    export_settings['current_paths'][k]['additional'] = []
                                if path_['path'] != export_settings['current_paths'][k]['path']:
                                    export_settings['current_paths'][k]['additional'].append(path_['path'])
                            else:
                                export_settings['current_paths'][k] = path_
                    export_settings['current_texture_transform'] = {}

            # We have all data for returning the Extension
            return Extension('KHR_materials_iridescence', iridescence_extension, False), uvmap_infos, udim_infos

        else:
            pass  # No texture found

    fac, path = get_factor_from_socket(iridescence_thickness_socket, kind='VALUE')
    if fac != 400.0:
        iridescence_extension['iridescenceThicknessMaximum'] = fac

        # Iridescence Thickness Maximum storing path for KHR_animation_pointer
        path_ = {}
        path_['length'] = 1
        path_['path'] = "/materials/XXX/extensions/KHR_materials_iridescence/iridescenceThicknessMaximum"
        export_settings['current_paths'][path] = path_

    # Iridescence Thickness Minimum
    if is_texture_iridescence is True:
        # Minimum is only used when there is a texture

        # No texture support here, only number
        fac, path = get_factor_from_socket(iridescence_thickness_minimum_socket, kind='VALUE')
        if fac != 100.0:
            iridescence_extension['iridescenceThicknessMinimum'] = fac

        # Iridescence Thickness Minimum storing path for KHR_animation_pointer
        path_ = {}
        path_['length'] = 1
        path_['path'] = "/materials/XXX/extensions/KHR_materials_iridescence/iridescenceThicknessMinimum"
        export_settings['current_paths'][path] = path_

    return Extension('KHR_materials_iridescence', iridescence_extension, False), uvmap_infos, udim_infos



# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
import dataclasses
import logging
import time
import urllib.parse
from pathlib import Path

from . import listing_downloader

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class CLIArguments:
    """Parsed command-line arguments."""

    url: str


def cli_main(arguments_raw: argparse.Namespace) -> None:
    """Generate the index for the passed-on-the-CLI asset library path."""

    # Parse CLI arguments.
    arguments = _parse_cli_args(arguments_raw)

    base_path = Path(".").resolve() / "_asset_download_location"  # TODO: be sensible.

    is_done = False

    def on_done_callback(_: listing_downloader.RemoteAssetListingDownloader) -> None:
        nonlocal is_done

        is_done = True

    downloader = listing_downloader.RemoteAssetListingDownloader(
        arguments.url,
        base_path,
        lambda *args: None,
        on_done_callback)
    downloader.download_and_process()

    while not is_done:
        # Ordinarily Blender's timer system will call the right method. But
        # because this is intended to run headless, and we're blocking the main
        # thread here, that doesn't happen.
        downloader.on_timer_event()
        time.sleep(downloader._DOWNLOAD_POLL_INTERVAL)

    print("Done!")


# Ignore the type of the `subparsers` argument, because there doesn't seem
# to be a way to make both static mypy and the runtime Python happy at the
# same time.


def add_cli_parser(subparsers: argparse._SubParsersAction) -> None:  # type: ignore[type-arg]
    """Add argparser for this subcommand."""

    parser = subparsers.add_parser("download", help="Download and parse a remote asset library index")
    parser.set_defaults(func=cli_main)

    parser.add_argument(
        "url",
        type=str,
        help="""URL of the remote asset library""",
    )


def _parse_cli_args(arguments_raw: argparse.Namespace) -> CLIArguments:
    """Make sure the passed arguments are valid."""

    try:
        urllib.parse.urlparse(arguments_raw.url)
    except ValueError as ex:
        logger.error("invalid URL specified: {}".format(ex))

    arguments = CLIArguments(
        url=arguments_raw.url,
    )

    return arguments


class APIVersionError(Exception):
    """Raised when none of the API versions declared by a remote asset library are supported by Blender."""



# SPDX-FileCopyrightText: 2018-2021 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

class MutatingArgument:
    """simple wrapper to pass a value by reference"""

    def __init__(self, value):
        self.value = value


def import_user_extensions(hook_name, gltf, *args):
    for extension in gltf.import_user_extensions:
        hook = getattr(extension, hook_name, None)
        if hook is not None:
            try:
                hook(*args, gltf)
            except Exception as e:
                if getattr(extension, 'is_critical', False):
                    gltf.log.error(
                        "Critical extension hook " +
                        hook_name +
                        " fails on " +
                        extension.__module__ +
                        ": " +
                        str(e),
                        popup=True)
                    raise RuntimeError("Import aborted due to critical extension failure") from e
                else:
                    gltf.log.error(hook_name, "fails on", extension)
                    gltf.log.error(str(e))



# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from itertools import batched

from . import blender_asset_library_openapi as api_models


def paginate_asset_list(
        assets: list[api_models.AssetV1],
        files: list[api_models.FileV1],
        num_assets_per_page: int = 0,
) -> list[api_models.AssetLibraryIndexPageV1]:
    """Return a list of asset pages.

    Each page is no longer than `num_assets_per_page` long. If zero, all assets
    are put in the same page.

    The files listed in each page are determined by the assets on that page.
    This means that it's possible for multiple pages to list the same file; this
    occurs when that file contains multiple assets, spread across multiple pages.
    """

    # Files are sorted to ensure the generated file is stable (i.e. regenerating produces the same file, and
    # inserting/removing files produce a small diff).
    def file_sort_key(file: api_models.FileV1) -> str:
        return file.path

    if not num_assets_per_page:
        return [api_models.AssetLibraryIndexPageV1(
            asset_count=len(assets),
            assets=assets,
            file_count=len(files),
            files=sorted(files, key=file_sort_key),
        )]

    pages = []
    for asset_batch in batched(assets, num_assets_per_page):
        used_file_paths = {
            file
            for asset in asset_batch
            for file in asset.files
        }
        file_batch = [file for file in files
                      if file.path in used_file_paths]
        file_batch.sort(key=file_sort_key)

        page = api_models.AssetLibraryIndexPageV1(
            asset_count=len(asset_batch),
            assets=list(asset_batch),
            file_count=len(file_batch),
            files=file_batch,
        )
        pages.append(page)
    return pages



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2012-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# Pre-process right-to-left languages.
# You can use it either standalone, or through import_po_from_branches or
# update_trunk.
#
# Notes: This has been tested on Linux, not 100% it will work nicely on
#        Windows or OsX.
#        This uses ctypes, as there is no py3 binding for fribidi currently.
#        This implies you only need the compiled C library to run it.
#        Finally, note that it handles some formatting/escape codes (like
#        \", %s, %x12, %.4f, etc.), protecting them from ugly (evil) fribidi,
#        which seems completely unaware of such things (as unicode is...).

import ctypes
import re


# define FRIBIDI_MASK_NEUTRAL    0x00000040L /* Is neutral */
FRIBIDI_PAR_ON = 0x00000040


# define FRIBIDI_FLAG_SHAPE_MIRRORING    0x00000001
# define FRIBIDI_FLAG_REORDER_NSM    0x00000002

# define FRIBIDI_FLAG_SHAPE_ARAB_PRES    0x00000100
# define FRIBIDI_FLAG_SHAPE_ARAB_LIGA    0x00000200
# define FRIBIDI_FLAG_SHAPE_ARAB_CONSOLE 0x00000400

# define FRIBIDI_FLAG_REMOVE_BIDI    0x00010000
# define FRIBIDI_FLAG_REMOVE_JOINING 0x00020000
# define FRIBIDI_FLAG_REMOVE_SPECIALS    0x00040000

# define FRIBIDI_FLAGS_DEFAULT       ( \
#   FRIBIDI_FLAG_SHAPE_MIRRORING    | \
#   FRIBIDI_FLAG_REORDER_NSM    | \
#   FRIBIDI_FLAG_REMOVE_SPECIALS    )

# define FRIBIDI_FLAGS_ARABIC        ( \
#   FRIBIDI_FLAG_SHAPE_ARAB_PRES    | \
#   FRIBIDI_FLAG_SHAPE_ARAB_LIGA    )

FRIBIDI_FLAG_SHAPE_MIRRORING = 0x00000001
FRIBIDI_FLAG_REORDER_NSM = 0x00000002
FRIBIDI_FLAG_REMOVE_SPECIALS = 0x00040000

FRIBIDI_FLAG_SHAPE_ARAB_PRES = 0x00000100
FRIBIDI_FLAG_SHAPE_ARAB_LIGA = 0x00000200

FRIBIDI_FLAGS_DEFAULT = FRIBIDI_FLAG_SHAPE_MIRRORING | FRIBIDI_FLAG_REORDER_NSM | FRIBIDI_FLAG_REMOVE_SPECIALS

FRIBIDI_FLAGS_ARABIC = FRIBIDI_FLAG_SHAPE_ARAB_PRES | FRIBIDI_FLAG_SHAPE_ARAB_LIGA


MENU_DETECT_REGEX = re.compile("%x\\d+\\|")


##### Kernel processing functions. #####
def protect_format_seq(msg):
    """
    Find some specific escaping/formatting sequences (like \", %s, etc.,
    and protect them from any modification!

    NOTE: This is not covering all exotic 'printf' formatting cases!
    It also only covers the minimal `{}` syntax for the modern `format` syntax.
    """
#    LRM = "\u200E"
#    RLM = "\u200F"
    LRE = "\u202A"
#    RLE = "\u202B"
    PDF = "\u202C"
    LRO = "\u202D"
#    RLO = "\u202E"
    # uctrl = {LRE, RLE, PDF, LRO, RLO}

    # 'printf' format, from https://cplusplus.com/reference/cstdio/printf/
    printf_format_flags = set("-+ #0")
    printf_format_widthprec = set(".0123456789")  # For width and precision.
    printf_format_datasize = set("hljztL")
    printf_format_codes = set("diuoxXfFeEgGaAcsp")
    # 'fmt::format' (and Python 'format()'),
    # see https://fmt.dev/12.0/syntax/ and https://docs.python.org/3.13/library/string.html#formatstrings
    fmt_format_widthprec = set(".0123456789")  # For width and precision.
    fmt_format_codes = set("aAbBcdeEfFgGnopsxX?%")

    if not msg:
        return msg
    elif MENU_DETECT_REGEX.search(msg):
        # An ugly "menu" message, just force it whole LRE if not yet done.
        if msg[0] not in {LRE, LRO}:
            msg = LRE + msg

    idx = 0
    ret = []
    ln = len(msg)
    while idx < ln:
        dlt = 1
#        # If we find a control char, skip any additional protection!
#        if msg[idx] in uctrl:
#            ret.append(msg[idx:])
#            break
        # \", \' or \\
        if idx < (ln - 1) and msg[idx] == '\\' and msg[idx + 1] in "\"\'\\":
            dlt = 2
        elif idx < (ln - 1) and msg[idx] == '{':
            # The whole 'format' syntax...
            # Coverage of this one is still fairly limited and basic currently.
            # TODO: support more of the 'format' mini-language (and check how much fmt::format matches with Python's).
            orig_dlt = dlt
            valid_format = False

            # {3}, {scale} (positional indicator or named reference)
            while (idx + dlt) < ln and msg[idx + dlt].isalnum():
                dlt += 1
            if (idx + dlt) < ln and msg[idx + dlt] == ":":
                dlt += 1
            # {:.4}, {:6d}, ...
            while (idx + dlt) < ln and msg[idx + dlt] in fmt_format_widthprec:
                dlt += 1
            # {:f}, {:s}, ...
            while (idx + dlt) < ln and msg[idx + dlt] in fmt_format_codes:
                dlt += 1
            if (idx + dlt) < ln and msg[idx + dlt] == "}":
                dlt += 1
                valid_format = True

            if not valid_format:
                dlt = orig_dlt
        # %%
        elif idx < (ln - 1) and msg[idx] == '%' and msg[idx + 1] == '%':
            dlt = 2
        elif idx < (ln - 1) and msg[idx] == '%':
            # The whole 'printf' syntax...
            # Not fully covering the format, but most of it, and should cover all Blender usages.
            orig_dlt = dlt
            valid_format = False

            # `%x12|` - What is this for actually?
            if idx < (ln - 2) and msg[idx + 1] in "x" and msg[idx + 2] in printf_format_widthprec:
                dlt = 2
                while (idx + dlt) < ln and msg[idx + dlt] in printf_format_widthprec:
                    dlt += 1
                if (idx + dlt) < ln and msg[idx + dlt] == '|':
                    dlt += 1
                    valid_format = True
            else:
                # `%+d, %-40s`, ...
                while (idx + dlt) < ln and msg[idx + dlt] in printf_format_flags:
                    dlt += 1
                # `%.4f, %6d`, ...
                while (idx + dlt) < ln and msg[idx + dlt] in printf_format_widthprec:
                    dlt += 1
                # `%lld, %zu`, ...
                while (idx + dlt) < ln and msg[idx + dlt] in printf_format_datasize:
                    dlt += 1
                # `%s, %d`, ...
                if (idx + dlt) < ln and msg[idx + dlt] in printf_format_codes:
                    dlt += 1
                    valid_format = True

            if not valid_format:
                dlt = orig_dlt

        if dlt > 1:
            ret.append(LRE)
        ret += msg[idx:idx + dlt]
        idx += dlt
        if dlt > 1:
            ret.append(PDF)

    return "".join(ret)


def log2vis(msgs, settings):
    """
    Globally mimics deprecated fribidi_log2vis.
    msgs should be an iterable of messages to RTL-process.
    """
    fbd = ctypes.CDLL(settings.FRIBIDI_LIB)

    for msg in msgs:
        msg = protect_format_seq(msg)

        fbc_str = ctypes.create_unicode_buffer(msg)
        ln = len(fbc_str) - 1
#        print(fbc_str.value, ln)
        btypes = (ctypes.c_int * ln)()
        embed_lvl = (ctypes.c_uint8 * ln)()
        pbase_dir = ctypes.c_int(FRIBIDI_PAR_ON)
        jtypes = (ctypes.c_uint8 * ln)()
        flags = FRIBIDI_FLAGS_DEFAULT | FRIBIDI_FLAGS_ARABIC

        # Find out direction of each char.
        fbd.fribidi_get_bidi_types(fbc_str, ln, ctypes.byref(btypes))

#        print(*btypes)

        fbd.fribidi_get_par_embedding_levels(
            btypes, ln,
            ctypes.byref(pbase_dir),
            embed_lvl,
        )

#        print(*embed_lvl)

        # Joinings for arabic chars.
        fbd.fribidi_get_joining_types(fbc_str, ln, jtypes)
#        print(*jtypes)
        fbd.fribidi_join_arabic(btypes, ln, embed_lvl, jtypes)
#        print(*jtypes)

        # Final Shaping!
        fbd.fribidi_shape(flags, embed_lvl, ln, jtypes, fbc_str)

#        print(fbc_str.value)
#        print(*(ord(c) for c in fbc_str))
        # And now, the reordering.
        # Note that here, we expect a single line, so no need to do
        # fancy things...
        fbd.fribidi_reorder_line(flags, btypes, ln, 0, pbase_dir, embed_lvl,
                                 fbc_str, None)
#        print(fbc_str.value)
#        print(*(ord(c) for c in fbc_str))

        yield fbc_str.value



# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

__all__ = [
    "download_asset_file",
    "downloader_status",
    "DownloadStatus",
]

from collections.abc import Callable
import dataclasses
import enum
import logging
import urllib.parse
from pathlib import Path

import bpy

from _bpy_internal.http import downloader as http_dl
from _bpy_internal.assets.remote_library.listing_downloader import RemoteAssetListingLocator
from _bpy_internal.assets.remote_library import hashing

logger = logging.getLogger(__name__)

# Preview images will NOT be downloaded if they already exist on disk AND their
# timestamp is younger than this age.
PREVIEW_DOWNLOAD_AGE_THRESHOLD_SEC = 7 * 24 * 3600  # 1 week

_asset_downloaders: dict[str, AssetDownloader] = {}
_preview_downloaders: dict[str, AssetDownloader] = {}


def download_asset_file(
        asset_library_url: str,
        asset_library_local_path: Path,
        asset_url: str,
        asset_hash: str,
        save_to: Path) -> str:
    """Download an asset file to a file on disk.

    :param asset_library_url: Root URL of the remote asset library. Used as an
        identifier of this library (to create a downloader per library), as well
        as for resolving relative URLs.

    :param asset_library_local_path: Root path of the local asset cache. Used to
        resolve relative `save_to` paths, but also to find the HTTP metadata
        cache for this asset library (for conditional downloads).

    :param asset_url: the URL to download. Can be absolute or relative to the
        asset library URL. If it is an empty string, the `save_to` path is used
        as the URL.
    :param asset_hash: the hash of the asset file, will be appended to the URL.

    :param save_to: the path on disk where to download to. While the download is
        pending, ".part" will be appended to the filename. When the download
        finishes successfully, it is renamed to the final path.

    :returns: the final URL that was queued for downloading.
    """
    try:
        downloader = _asset_downloaders[asset_library_url]
        assert downloader.local_path == asset_library_local_path, (
            "This code assumes that remote asset libraries do not move on the local disk"
        )
    except KeyError:
        downloader = AssetDownloader(
            asset_library_url,
            asset_library_local_path,
            reporter=AssetReporter(asset_library_url=asset_library_url),
            on_queue_empty_callback=on_asset_download_queue_empty,
            queue_side=http_dl.QueueSide.BACK,  # FIFO queue.
            num_parallel_downloads=1,
        )
        downloader.start()
        _asset_downloaders[asset_library_url] = downloader

    # Construct the URL if not given explicitly.
    if not asset_url:
        if save_to.is_absolute():
            relative_path = save_to.relative_to(asset_library_local_path)
        else:
            relative_path = save_to
        asset_url = urllib.parse.quote(relative_path.as_posix())

    # Include the hash in the URL, and download the asset.
    download_url = hashing.url((asset_url, asset_hash))
    full_url = downloader.download_asset_file(download_url, save_to)
    return full_url


def download_preview(
        asset_library_url: str,
        asset_library_local_path: Path,
        preview_url: str,
        preview_hash: str,
        dst_filepath: Path) -> None:
    """Download an asset preview to a file on disk.

    :param asset_library_url: Root URL of the remote asset library. Used as an
        identifier of this library (to create a downloader per library), as well
        as for resolving relative URLs.

    :param asset_library_local_path: Root path of the local asset cache. Used to
        resolve relative `save_to` paths, but also to find the HTTP metadata
        cache for this asset library (for conditional downloads).

    :param preview_url: the URL to download. Can be absolute or relative.
    :param preview_hash: the hash of the thumbnail, will be appended to the URL.

    :param dst_filepath: the path on disk where to download to. While the
        download is pending, ".part" will be appended to the filename. When the
        download finishes successfully, it is renamed to the final path.
    """
    import time

    # Check if the file exists and is new enough. If it is, don't bother the server.
    try:
        stat = dst_filepath.stat()
    except FileNotFoundError:
        pass  # Fine, something new to download.
    else:
        # File exists, let's see if it's young enough to use as-is.
        age_in_seconds = time.time() - stat.st_mtime
        if age_in_seconds < PREVIEW_DOWNLOAD_AGE_THRESHOLD_SEC:
            # The local file is still fresh, just pretend we just downloaded it.
            bpy.types.WindowManager.asset_library_status_ping_loaded_new_preview(str(dst_filepath))
            return

    try:
        downloader = _preview_downloaders[asset_library_url]
        assert downloader.local_path == asset_library_local_path, (
            "This code assumes that remote asset libraries do not move on the local disk"
        )
    except KeyError:
        downloader = AssetDownloader(
            asset_library_url,
            asset_library_local_path,
            reporter=PreviewReporter(),
            on_queue_empty_callback=None,
            queue_side=http_dl.QueueSide.FRONT,  # LIFO queue.
            num_parallel_downloads=5,
        )
        downloader.start()
        _preview_downloaders[asset_library_url] = downloader

    # Include the hash in the URL, and download the preview.
    download_url = hashing.url((preview_url, preview_hash))
    downloader.download_asset_file(download_url, dst_filepath)


def cancel_download(asset_library_url: str, full_asset_url: str) -> None:
    """Cancel a running/queued asset download.

    Cancelling a URL that has already been fully downloaded, or one that was never
    queued is a no-op.

    :param asset_library_url: Root URL of the remote asset library. Used as an
        identifier of this library (to create a downloader per library).
        Contrary to the download function, this is NOT used to resolve relative
        URLs.
    :param full_asset_url: the URL that's queued for download. MUST be the final
        URL as returned by download_asset_file().
    """

    try:
        downloader = _asset_downloaders[asset_library_url]
    except KeyError:
        # No downloader could mean that the cancel came in just a millisecond
        # too late, and the download was already finished.
        return

    downloader.cancel_download(full_asset_url)


def cancel_download_all_assets() -> None:
    """Cancel all active/queued downloads of all assets.

    This shuts down all asset downloaders, effectively cancelling all their downloads.
    """

    for downloader in _asset_downloaders.values():
        downloader.cancel_and_shutdown()


def downloader_status(asset_library_url: str) -> DownloadStatus:
    """Returns the asset downloader status.

    Raises a KeyError if there never was a downloader for this URL.
    """
    return _asset_downloaders[asset_library_url].status


def on_asset_download_queue_empty() -> None:
    """Called by the asset downloader when its download queue emptied."""
    if any_asset_downloading():
        return
    bpy.types.WindowManager.asset_library_status_ping_finished_download_queue()


def any_asset_downloading() -> bool:
    """Returns true if there is any downloader currently downloading assets."""
    return any(
        downloader.status == DownloadStatus.DOWNLOADING
        for downloader in _asset_downloaders.values()
    )


class DownloadStatus(enum.Enum):
    IDLE = 'idle'
    DOWNLOADING = 'downloading'

    FINISHED = 'finished'
    """The downloader has downloaded everything that was queued.

    Note: this does NOT mean that all downloads were perfect. It just means that
    there were no exceptions raised.
    """

    FAILED = 'failed'
    """Unexpected exceptions occurred."""

    CANCELLED = 'cancelled'
    """There still were pending downloads when the downloader shut down."""


class AssetDownloader:
    """Downloader for asset files & their thumbnails."""

    _locator: RemoteAssetListingLocator
    _bg_downloader: http_dl.BackgroundDownloader | None
    _reporter: http_dl.DownloadReporter
    _num_assets_pending: int

    type QueueEmptyCallback = Callable[[], None]
    _on_queue_empty_callback: QueueEmptyCallback | None
    """Called when the download queue became empty."""

    _status: DownloadStatus
    _error_message: str
    """An error message to show to the user.

    Should be set on errors to communicate a message to users. Calling report()
    with 'ERROR' as the level will set this to the given message.
    """

    _DOWNLOAD_POLL_INTERVAL: float = 0.01
    """How often the background download process is polled, in seconds.

    Each 'poll' involves sending queued messages back & forth between the main
    Blender process and the background download process.
    """

    _HTTP_METHOD = "GET"

    _queue_side: http_dl.QueueSide
    """Previews are queued at the front (LIFO), and assets at the back (FIFO)."""

    def __init__(
        self,
        remote_url: str,
        local_path: Path | str,
        *,
        reporter: http_dl.DownloadReporter,
        on_queue_empty_callback: QueueEmptyCallback | None,
        queue_side: http_dl.QueueSide,
        num_parallel_downloads: int,
    ) -> None:
        """Create a downloader for assets of a specific asset library.

        :param remote_url: Base URL of the remote asset library server.

        :param local_path: The directory to download the index files to.

        :param on_download_done_callback: called with one parameter (this
            AssetDownloader) when a file finished downloading and was put
            in its final location, ready to be picked up by the asset system.
        """
        self._locator = RemoteAssetListingLocator(remote_url, local_path)
        self._num_assets_pending = 0
        self._reporter = reporter
        self._on_queue_empty_callback = on_queue_empty_callback

        self._status = DownloadStatus.IDLE
        self._error_message = ""

        # Work around a limitation of Blender, see bug report #139720 for details.
        self.on_timer_event = self.on_timer_event  # type: ignore[method-assign]

        self._downloader_options = http_dl.DownloaderOptions(
            metadata_provider=http_dl.MetadataProviderFilesystem(
                cache_location=self._locator.http_metadata_cache_location,
            ),
            timeout=300,
            http_headers={
                'X-Blender': "{:d}.{:d}".format(*bpy.app.version),
            },
            num_parallel_downloads=num_parallel_downloads,
        )

        self._bg_downloader = None
        self._queue_side = queue_side

    def _create_bg_downloader(self) -> None:
        self._bg_downloader = http_dl.BackgroundDownloader(
            options=self._downloader_options,
            on_callback_error=self._on_callback_error,
        )

        # These are called in order. Doing things this way ensures that self._reporter.download_finished() is called for
        # every individual download, and after that our own function is called. That means that the
        # self._on_queue_empty_callback() function is called _after_ the individual downloads.
        #
        # Swapping this order would mean self._on_queue_empty_callback() is called _before_ the last call to
        # self._reporter.download_finished(), which would be confusing.
        self._bg_downloader.add_reporter(self._reporter)
        self._bg_downloader.add_reporter(self)

    def __repr__(self) -> str:
        return "{!s}(remote_url={!r}, local_path={!r})".format(
            type(self),
            self._locator.remote_url,
            self._locator.local_path,
        )

    def start(self) -> None:
        """Start the background process."""
        if not self._bg_downloader:
            self._create_bg_downloader()
            assert self._bg_downloader
        self._bg_downloader.start()

        # Register the timer for periodic message passing between the main and
        # background processes.
        if not bpy.app.timers.is_registered(self.on_timer_event):
            bpy.app.timers.register(
                self.on_timer_event,
                first_interval=self._DOWNLOAD_POLL_INTERVAL,
                persistent=True,
            )
            # Double-check the registration worked, see #139720 for details.
            assert bpy.app.timers.is_registered(self.on_timer_event)

    def download_asset_file(self, asset_url: str, save_to: Path) -> str:
        """Download an asset or preview file to a local file.

        Returns the URL that was queued. This is different than the given URL
        when the latter is relative.
        """

        # If the downloader was shut down, start it up again.
        if not self._bg_downloader:
            self.start()

        self._status = DownloadStatus.DOWNLOADING
        url = self._queue_download(asset_url, save_to)
        return url

    def cancel_download(self, full_asset_url: str) -> None:
        """Cancel downloading a URL.

        If the URL was never queued, or it has already been downloaded,
        this is a no-op.
        """
        if not self._bg_downloader:
            return

        logger.info("cancelling download of %s", full_asset_url)
        http_req_descr = http_dl.RequestDescription(self._HTTP_METHOD, full_asset_url)
        self._bg_downloader.cancel_download(http_req_descr)

    def _shutdown_if_done(self) -> None:
        if self._num_assets_pending > 0:
            return

        is_done = self._bg_downloader is None or self._bg_downloader.all_downloads_done
        if not is_done:
            return

        # Done downloading everything, let's shut down.
        self._status = DownloadStatus.FINISHED

        if self._on_queue_empty_callback is not None:
            # Call the callback _after_ setting the status, so that when
            # Blender is pinged about this, it can see it's finished.
            self._on_queue_empty_callback()

        # TODO: delay this for a few minutes, so that we don't need a new
        # background process for every asset.
        self.shutdown()

    def _on_callback_error(
            self,
            http_req_descr: http_dl.RequestDescription,
            local_file: Path,
            exception: Exception) -> None:
        logger.exception(
            "exception while handling downloaded file ({!r}, saved to {!r})".format(
                http_req_descr, local_file))
        self.report({'ERROR'}, "Resource download had an issue, download aborted")
        self._status = DownloadStatus.FAILED
        self.shutdown()

    def _queue_download(self, asset_url: str, download_to_path: Path | str) -> str:
        """Queue up this download.

        Returns the URL of the download, and the path to which it will be downloaded.
        """
        remote_url = urllib.parse.urljoin(self._locator.remote_url, asset_url)
        download_to_path = self._locator.local_path / download_to_path

        # Safety measure: refuse to download a file into the listing directory.
        if self._locator.is_system_path(download_to_path):
            raise ValueError(
                ("Asset at {!s} wants to be downloaded to {!s}, which would overwrite local asset system files. " +
                 "Notify the owner of the asset library about this.").format(
                    remote_url,
                    download_to_path))

        logger.info("downloading %s to %s", remote_url, download_to_path)

        assert self._bg_downloader, "downloads can only be queued when the bgdownloader is available"
        request_descr = self._bg_downloader.queue_download(
            remote_url,
            download_to_path,
            http_method=self._HTTP_METHOD,
            queue_side=self._queue_side,
        )
        return request_descr.url

    # TODO: implement this in a more useful way:
    def report(self, level: set[str], message: str) -> None:
        # logger.info("Report: {:s}: {:s}".format("/".join(level), message))
        if 'ERROR' in level:
            self._error_message = message

    def cancel_and_shutdown(self) -> None:
        """Cancel all downloads and shut down the background downloader."""

        # Only set to 'Cancelled' if the downloader was still downloading.
        if self._status == DownloadStatus.DOWNLOADING:
            if self._bg_downloader and self._bg_downloader.num_pending_downloads > 0:
                self._status = DownloadStatus.CANCELLED
            else:
                self._status = DownloadStatus.FINISHED

        # The downloads themselves don't have to be explicitly cancelled,
        # shutting down the downloader will do that implicitly.
        self.shutdown()

        # By now there is no more queue, so just treat it as 'empty' and let Blender know no downloads will happen any
        # more (at least not by this downloader).
        if self._on_queue_empty_callback is not None:
            # Call the callback _after_ setting the status, so that when
            # Blender is pinged about this, it can see it's finished.
            self._on_queue_empty_callback()

    def shutdown(self) -> None:
        """Stop the background downloader and call the 'done' callback."""

        # The timer is no longer necessary, the bg_downloader.shutdown() call
        # takes care of the last queued messages.
        if bpy.app.timers.is_registered(self.on_timer_event):
            bpy.app.timers.unregister(self.on_timer_event)

        try:
            if not self._bg_downloader:
                return

            # Only report if this is actually triggering a shutdown. If that was
            # already triggered somehow, don't bother.
            if not self._bg_downloader.is_shutdown_requested:
                # It may be tempting to call self.report(...) here, and report on the
                # cancellation. However, this should be done by the caller, when they know
                # of the reason of the cancellation and thus can provide more info.
                num_pending = self._bg_downloader.num_pending_downloads
                if num_pending:
                    logger.warning("Shutting down background downloader, %d downloads pending", num_pending)

            self._bg_downloader.shutdown()
        finally:
            # Regardless of whether the shutdown had some issues, the timer has
            # been unregistered, so there will be no more message handling, and
            # so for all intents and purposes, the downloader is done.
            self._bg_downloader = None

    def on_timer_event(self) -> float:
        assert self._bg_downloader, "timer events should only come in while the bgdownloader is available"

        try:
            self._bg_downloader.update()
        except http_dl.BackgroundProcessNotRunningError:
            logger.error("Background downloader subprocess died, aborting.")
            self._status = DownloadStatus.FAILED
            self.shutdown()
            return 0  # Deactivate the timer.
        except Exception:
            logger.exception(
                "Unexpected error downloading remote asset library ilisting from %s to %s",
                self._locator.remote_url,
                self._locator.local_path)

        # Automatically switch between IDLE and DOWNLOADING, but never overwrite
        # FAILED or FINISHED_SUCCESFULLY.
        if self._status in {DownloadStatus.DOWNLOADING, DownloadStatus.IDLE}:
            if self._bg_downloader.num_pending_downloads > 0:
                self._status = DownloadStatus.DOWNLOADING
            else:
                self._status = DownloadStatus.IDLE

        return self._DOWNLOAD_POLL_INTERVAL

    @property
    def remote_url(self) -> str:
        return self._locator.remote_url

    @property
    def local_path(self) -> Path:
        return self._locator.local_path

    @property
    def status(self) -> DownloadStatus:
        return self._status

    @property
    def error_message(self) -> str:
        return self._error_message

    # Below here: http_dl.DownloadReporter protocol functions:

    def download_starts(self, http_req_descr: http_dl.RequestDescription) -> None:
        pass

    def already_downloaded(
        self,
        http_req_descr: http_dl.RequestDescription,
        local_file: Path,
    ) -> None:
        self._shutdown_if_done()

    def download_error(
        self,
        http_req_descr: http_dl.RequestDescription,
        local_file: Path,
        error: Exception,
    ) -> None:
        if isinstance(error, http_dl.DownloadCancelled):
            # Cancelling a download should cancel all queued-up downloads.
            if self._num_assets_pending:
                self.report({'WARNING'}, "Cancelled {} pending download".format(self._num_assets_pending))
            logger.warning("Download cancelled: %s", http_req_descr)
            self._status = DownloadStatus.FAILED
            self.shutdown()
            return

        self._shutdown_if_done()

    def download_progress(
        self,
        http_req_descr: http_dl.RequestDescription,
        progress: http_dl.DownloadProgress,
    ) -> None:
        pass

    def download_finished(
        self,
        http_req_descr: http_dl.RequestDescription,
        local_file: Path,
    ) -> None:
        self._shutdown_if_done()


@dataclasses.dataclass
class AssetReporter:
    """Implementation of the http_dl.DownloadReporter protocol."""

    asset_library_url: str

    def download_starts(self, http_req_descr: http_dl.RequestDescription) -> None:
        logger.debug("Download starting: %s", http_req_descr)

    def already_downloaded(
        self,
        http_req_descr: http_dl.RequestDescription,
        local_file: Path,
    ) -> None:
        logger.debug("Download unnecessary, file already downloaded: %s", http_req_descr.url)
        bpy.types.WindowManager.asset_library_status_ping_asset_file_succeeded(
            self.asset_library_url, http_req_descr.url, str(local_file))

    def download_error(
        self,
        http_req_descr: http_dl.RequestDescription,
        local_file: Path,
        error: Exception,
    ) -> None:
        logger.warning("Could not download file %s: %s", http_req_descr, error)
        bpy.types.WindowManager.asset_library_status_ping_asset_file_failed(
            self.asset_library_url, http_req_descr.url, str(local_file))

    def download_progress(
        self,
        http_req_descr: http_dl.RequestDescription,
        progress: http_dl.DownloadProgress,
    ) -> None:
        bpy.types.WindowManager.asset_library_status_ping_asset_file_progress(
            http_req_descr.url, progress.disk_bytes_written)

    def download_finished(
        self,
        http_req_descr: http_dl.RequestDescription,
        local_file: Path,
    ) -> None:
        logger.info("Download finished: %s to %s", http_req_descr, local_file)
        bpy.types.WindowManager.asset_library_status_ping_asset_file_succeeded(
            self.asset_library_url, http_req_descr.url, str(local_file))


@dataclasses.dataclass
class PreviewReporter:
    """Implementation of the http_dl.DownloadReporter protocol."""

    def download_starts(self, http_req_descr: http_dl.RequestDescription) -> None:
        logger.debug("Download starting: %s", http_req_descr)

    def already_downloaded(
        self,
        http_req_descr: http_dl.RequestDescription,
        local_file: Path,
    ) -> None:
        # This cannot check the content-type header (like download_finished() does), since
        # there likely is none in a '304 Not Modified' response.

        # Indicate to a future run that we just confirmed this file is still fresh.
        local_file.touch()

        # Poke Blender so it knows there's a thumbnail update. It shouldn't be necessary, but since it requested the
        # file for downloading, it may not have been aware it already existed. Better let it know.
        bpy.types.WindowManager.asset_library_status_ping_loaded_new_preview(str(local_file))

    def download_error(
        self,
        http_req_descr: http_dl.RequestDescription,
        local_file: Path,
        error: Exception,
    ) -> None:
        # TODO: create an empty file in the correct `.../_thumbs/failed` directory.

        # Poke Blender so it knows there's a thumbnail update.
        bpy.types.WindowManager.asset_library_status_ping_loaded_new_preview(str(local_file))

    def download_progress(
        self,
        http_req_descr: http_dl.RequestDescription,
        progress: http_dl.DownloadProgress,
    ) -> None:
        pass

    def download_finished(
        self,
        http_req_descr: http_dl.RequestDescription,
        local_file: Path,
    ) -> None:
        # Check whether the file was actually an image.
        assert http_req_descr.response_headers
        content_type = http_req_descr.response_headers.get('content-type', "")

        # Only check the content type if the server sends it back. Otherwise
        # just trust that it's valid. For example, when sending a `304 Not
        # Modified`, the server may actually skip the Content-Type header.
        if content_type and not content_type.startswith('image/'):
            logger.warning("Thumbnail URL %r has content type %r, expected an image",
                           http_req_descr.url, content_type)
            # TODO: mark as 'failed' so that this file isn't repeatedly
            # downloaded and rejected. For now I'll just keep the file
            # around, so that at least the time-stamping works to prevent
            # hammering the server.

        # Indicate to a future run that we just confirmed this file is still fresh.
        local_file.touch()

        # Poke Blender so it knows there's a thumbnail update.
        bpy.types.WindowManager.asset_library_status_ping_loaded_new_preview(str(local_file))



# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

"""Blender Online Asset Repository Listing Generator."""

__all__ = (
    'cli_main',
    'SCHEMA_VERSION',
)

import argparse
import dataclasses
import json
import logging
import sys
import urllib.parse
from pathlib import Path
from typing import Any

import cattrs.preconf.json

from . import hashing, listing_asset_catalogs, listing_common, json_parsing
from . import cli_listing_generator_asset_finder as asset_finder
from . import cli_listing_generator_pagination as pagination
from . import blender_asset_library_openapi as api_models

SCHEMA_VERSION = "1.0.0"
DEFAULT_METADATA = api_models.AssetLibraryMeta(
    api_versions={},  # Determined by cli_main().
    name="Your Asset Library",
    contact=api_models.Contact(
        name="Your Name",
        url="https://example.org/",
        email="example@example.org",
    ),
)

logger = logging.getLogger(__name__)

_converter = cattrs.preconf.json.JsonConverter(omit_if_default=True)


@dataclasses.dataclass
class CLIArguments:
    """Parsed commandline arguments."""

    repository: Path
    limit: int
    page_size: int


def cli_main(arguments_raw: argparse.Namespace) -> None:
    """Generate the index for the passed-on-the-CLI asset library path."""

    # Parse CLI arguments.
    arguments = _parse_cli_args(arguments_raw)

    # Read the top-level meta file first. If this already exists, an attempt
    # at parsing & upgrading it is performed. Better to do this (and stop on
    # errors) before diving into the assets themselves.
    meta_json_path = arguments.repository / listing_common.ASSET_TOP_METADATA_FILENAME
    toplevel_meta = _toplevel_meta_read(meta_json_path)

    # Find all .blend files.
    filepaths: list[Path] = []
    logger.info("Traversing %s", arguments.repository)
    for filepath in arguments.repository.rglob("*.blend"):
        filepaths.append(filepath)

    files_total = len(filepaths)
    logger.info(f"* {files_total} .blend files found.")

    limit = _total_files_to_process(arguments, files_total)

    # Find the assets in the blend files.
    logger.info("Parsing the files...")
    assets: list[api_models.AssetV1] = []
    files: list[api_models.FileV1] = []

    for i, filepath in enumerate(filepaths[:limit]):
        logger.info(f"* {i + 1}/{limit}: {filepath.relative_to(arguments.repository)}")

        bfile_info, assets_in_file = asset_finder.list_assets(filepath, arguments.repository)
        if not assets_in_file:
            continue

        assets.extend(assets_in_file)
        files.append(bfile_info)

    _sort_assets(assets)

    # Write the listing index and the pages:
    asset_index_pages = pagination.paginate_asset_list(assets, files, arguments.page_size)
    index_path = _write_json_files(arguments, asset_index_pages)

    # Write the top-level meta file:
    api_version_key = "v{:d}".format(listing_common.API_VERSION)
    index_relpath: Path = index_path.relative_to(arguments.repository)
    toplevel_meta.api_versions[api_version_key] = api_models.URLWithHash(
        url=urllib.parse.quote(index_relpath.as_posix()),
        hash=hashing.hash_file(index_path),
    )
    _save_json(toplevel_meta, meta_json_path)


def _toplevel_meta_read(meta_json_path: Path) -> api_models.AssetLibraryMeta:
    try:
        metadata = _toplevel_metadata(meta_json_path)
    except (json.JSONDecodeError, cattrs.errors.ClassValidationError) as ex:
        msg = "Metadata file {} could not be parsed: {}"
        logger.error(msg.format(meta_json_path, ex))
        raise SystemExit(1) from None
    return metadata


def _sort_assets(assets: list[api_models.AssetV1]) -> None:
    """Sorts the assets in-place.

    Sorting helps to get the generated listing stable, so that a diff between
    two runs of the generator is as clean as possible.
    """

    # Sort assets by their primary filename first. This places related assets together, and minimizes the repeats of the
    # same file across multiple listing pages.
    def sort_key(asset: api_models.AssetV1) -> tuple[str, str, str]:
        if asset.files:
            first_file = asset.files[0].lower()
        else:
            first_file = ""
        return (first_file, asset.id_type.lower(), asset.name.lower())

    assets.sort(key=sort_key)


def _write_json_files(
    arguments: CLIArguments,
    asset_index_pages: list[api_models.AssetLibraryIndexPageV1],
) -> Path:
    """Write the asset listing page files and the index file.

    :returns: the path of the index file.
    """
    outdir_root = arguments.repository
    outdir_versioned = outdir_root / listing_common.API_VERSIONED_SUBDIR

    # Remove old pages, in case the number of assets per page was increased and
    # so less page files are needed.
    existing_pages = outdir_versioned.glob("assets-*.json")
    for filepath in existing_pages:
        filepath.unlink()

    # Library Index Page /_v1/assets-{page}.json
    #
    # Note that these paths are determined by the generator, and their URLs are
    # listed explicitly in the index file, so there is no need to have those in
    # the listing_common.py file.
    page_infos: list[api_models.URLWithHash] = []
    for page_index, page in enumerate(asset_index_pages):
        page_relpath = listing_common.api_versioned(f"assets-{page_index:05}.json")
        page_abspath = outdir_root / page_relpath
        _save_json(page, page_abspath)

        page_infos.append(api_models.URLWithHash(
            url=urllib.parse.quote(page_relpath.as_posix()),
            hash=hashing.hash_file(page_abspath),
        ))

    # Library Index file /_v1/asset-index.json:
    total_asset_count = sum(page.asset_count for page in asset_index_pages)
    total_file_count = sum(page.file_count for page in asset_index_pages)
    asset_size_bytes = sum(file.size_in_bytes
                           for page in asset_index_pages
                           for file in page.files)
    asset_cats = listing_asset_catalogs.parse_catalogs(arguments.repository)
    index = api_models.AssetLibraryIndexV1(
        schema_version=SCHEMA_VERSION,
        asset_size_bytes=asset_size_bytes,
        asset_count=total_asset_count,
        file_count=total_file_count,
        pages=page_infos,
        catalogs=asset_cats,
    )
    index_path = outdir_versioned / listing_common.ASSET_INDEX_JSON_FILENAME
    _save_json(index, index_path)

    return index_path


def _save_json(model: Any, json_path: Path) -> None:
    as_json = _converter.dumps(model, indent=2)

    json_path.parent.mkdir(exist_ok=True, parents=True)

    logger.info("Writing %s", json_path)
    with json_path.open("wt") as json_file:
        json_file.write(as_json)


def _toplevel_metadata(json_path: Path) -> api_models.AssetLibraryMeta:
    """Construct the top-level metadata.

    Returns the metadata, or raises an exception (see json_parsing.ValidatingParser)
    if it is not valid JSON.

    Writing is considered safe, except when the file exists but does not contain
    valid JSON. In that case, it's better to warn about this and keep the file
    as-is, so that the user can either delete or fix it.
    """
    try:
        json_data = json_path.read_bytes()
    except IOError:
        # Ignore any read errors, as this likely means the file simply doesn't exist.
        return DEFAULT_METADATA

    parser = json_parsing.ValidatingParser()
    metadata = parser.parse_and_validate(api_models.AssetLibraryMeta, json_data)

    # Update the metadata to declare the API version for which we're going to
    # write the data.
    metadata.api_versions = DEFAULT_METADATA.api_versions.copy()

    return metadata


# Ignore the type of the `subparsers` argument, because there doesn't seem
# to be a way to make both static mypy and the runtime Python happy at the
# same time.
def add_cli_parser(subparsers: argparse._SubParsersAction) -> None:  # type: ignore[type-arg]
    """Add argparser for this subcommand."""

    parser = subparsers.add_parser("generate", help="Generate files necessary to serve an asset library")
    parser.set_defaults(func=cli_main)

    parser.add_argument(
        "repository",
        type=Path,
        help="""Asset repository folder""",
    )

    parser.add_argument(
        "--limit",
        "-l",
        metavar="NUM_BLEND_FILES",
        type=int,
        default=None,
        help="Limit the number of files to process",
    )

    parser.add_argument(
        "--page",
        "-p",
        metavar="ASSETS_PER_PAGE",
        type=int,
        default=1000,
        help="Number of assets per JSON file, set to 0 to disable pagination",
    )


def _parse_cli_args(arguments_raw: argparse.Namespace) -> CLIArguments:
    """Make sure the passed arguments are valid."""

    repository = arguments_raw.repository.absolute()
    if not repository.is_dir():
        print(f"Error: Repository specified is not a folder: {repository}")
        sys.exit(1)

    arguments = CLIArguments(
        repository=repository,
        limit=arguments_raw.limit or 0,
        page_size=arguments_raw.page or 0,
    )

    return arguments


def _total_files_to_process(arguments: CLIArguments, files_total: int) -> int:
    if not arguments.limit:
        return files_total

    return min(arguments.limit, files_total)



# SPDX-FileCopyrightText: 2026 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

import ctypes
import sys
import os
from .gltf2_io_gltf import ImportError
from ..com.library import dll_path


class MeshoptDecoder:
    """Meshopt decoder."""
    def __new__(cls, *args, **kwargs):
        raise RuntimeError("%s should not be instantiated" % cls)

    @staticmethod
    def find_library():
        """Find the Meshopt decoder library."""
        path = dll_path('bf_intern_meshopt_bridge', 'MeshOptimizer')
        if path is not None and path.exists() and path.is_file():
            return path
        else:
            raise ImportError("Meshopt decoder library not found at {}".format(path))

    @staticmethod
    def load_library(gltf):
        """Load the Meshopt decoder library."""
        if hasattr(gltf, 'meshopt_decoder'):
            return
        lib_path = MeshoptDecoder.find_library()
        try:
            lib = ctypes.CDLL(lib_path.resolve())
        except Exception as e:
            raise ImportError("Failed to load Meshopt decoder library: {}".format(e))

        gltf.meshopt_decoder = lib

        # Define type signatures for the decoder functions
        decode_funcs = [
            lib.decodeVertexBuffer,
            lib.decodeIndexBuffer,
            lib.decodeIndexSequence,
        ]
        for func in decode_funcs:
            func.restype = ctypes.c_int
            func.argtypes = [
                ctypes.c_void_p,  # void* destination
                ctypes.c_size_t,  # size_t count
                ctypes.c_size_t,  # size_t stride
                ctypes.POINTER(ctypes.c_ubyte),  # const unsigned char* buffer
                ctypes.c_size_t,  # size_t buffer_size
            ]

        filter_funcs = [
            lib.decodeFilterOct,
            lib.decodeFilterQuat,
            lib.decodeFilterExp,
        ]
        for func in filter_funcs:
            func.restype = None
            func.argtypes = [
                ctypes.c_void_p,  # void* buffer
                ctypes.c_size_t,  # size_t count
                ctypes.c_size_t,  # size_t stride
            ]

    @staticmethod
    def get_buffer_view(gltf, bufferview_index):
        """Decodes EXT/KHR_meshopt_compression buffer view."""
        # Check if already in cache
        if not hasattr(gltf, 'meshopt_cache'):
            gltf.meshopt_cache = {}
        if bufferview_index in gltf.meshopt_cache:
            return gltf.meshopt_cache[bufferview_index]

        bufview = gltf.data.buffer_views[bufferview_index]
        # TODO: make sure we detect before that we have EXT or KHR
        if 'EXT_meshopt_compression' in bufview.extensions:
            ext = bufview.extensions['EXT_meshopt_compression']
        else:
            ext = bufview.extensions['KHR_meshopt_compression']

        buffer_idx = ext['buffer']
        byte_length = ext['byteLength']
        byte_offset = ext.get('byteOffset', 0)
        byte_stride = ext['byteStride']
        count = ext['count']
        mode = ext['mode']
        filter = ext.get('filter', None)

        MeshoptDecoder.load_library(gltf)
        lib = gltf.meshopt_decoder

        # load buffer
        if buffer_idx not in gltf.buffers:
            gltf.load_buffer(buffer_idx)

        buffer = gltf.buffers[buffer_idx]
        buffer = buffer[byte_offset:byte_offset + byte_length]

        # Create output buffer
        output = memoryview(bytearray(count * byte_stride))
        dst_ptr = (ctypes.c_ubyte * len(output)).from_buffer(output)
        # TODO: this creates a copy of the buffer, we should be able to decode in place without copy
        buffer_ptr = (ctypes.c_ubyte * len(buffer)).from_buffer_copy(buffer)

        decode_func = {
            'ATTRIBUTES': lib.decodeVertexBuffer,
            'TRIANGLES': lib.decodeIndexBuffer,
            'INDICES': lib.decodeIndexSequence,
        }.get(mode)

        error_code = decode_func(
            dst_ptr,
            count,
            byte_stride,
            buffer_ptr,
            len(buffer),
        )
        if error_code != 0:
            raise ImportError("Meshopt decoding failed with error code {}".format(error_code))

        if mode == 'ATTRIBUTES' and filter is not None:
            filter_func = {
                'OCTAHEDRAL': lib.decodeFilterOct,
                'QUATERNION': lib.decodeFilterQuat,
                'EXPONENTIAL': lib.decodeFilterExp,
                'COLOR': lib.decodeFilterExp,
            }.get(filter)
            filter_func(dst_ptr, count, byte_stride)

        # Cache the decoded buffer view
        gltf.meshopt_cache[bufferview_index] = output
        return output



# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Generated by datamodel-codegen:
#   source filename:  blender_asset_library_openapi.yaml

from __future__ import annotations

from dataclasses import dataclass
from enum import StrEnum
from typing import Any


@dataclass
class Contact:
    """Owner / publisher of this asset library."""

    name: str
    url: str | None = None
    email: str | None = None


@dataclass
class URLWithHash:
    """Resource that's identified by a URL.

    The resource should be fetched by including the hash in the query
    string, like `GET {URL}?hash={HASH}`. Here `{HASH}` should _not_
    include the hash type. The purpose of including this on the URL is
    for cache busting, and thus the hash type is not relevant here.
    """

    url: str
    """URL of the page file."""

    hash: str
    """Hash of the resource obtained at that URL.

    This should be in the format "HASHTYPE:HASH-AS-HEX". Currently only
    the "SHA256" hash type is supported. Note that for dynamic API
    servers, which may perform a server-side filter on the data, the
    actual response may not have the same hash. Static servers send
    content that matches the hash.
    """


type AssetIDTypeV1 = str
"""Type of the Blender data-block.

This can be obtained via BPY with `datablock.id_type`. Any comparisons
should be done in a case-insensitive manner. Note that this list is just
a list of data-block types in Blender. This type being in this list does
not mean that Blender supports making this data-block an asset. It's
just here to ensure that if that changes, and more data-block types can
become assets, this schema doesn't need updating.
"""


class CustomPropertyTypeV1(StrEnum):
    """Type of IDProperty, see `eIDPropertyType` in `DNA_ID_enumms.h`.

    For now, type `ID` and `IDPARRAY` are not supported.
    """

    IDP_STRING = "IDP_STRING"
    IDP_INT = "IDP_INT"
    IDP_FLOAT = "IDP_FLOAT"
    IDP_ARRAY = "IDP_ARRAY"
    IDP_GROUP = "IDP_GROUP"
    IDP_DOUBLE = "IDP_DOUBLE"
    IDP_BOOL = "IDP_BOOL"


@dataclass
class AssetBlenderVersionsV1:
    """Minimum and (optionally) maximum versions of Blender that this asset
    should be shown in.

    This is a half-open interval: Blender shows the asset if `min <= blender < until`.
    """

    min: str
    """Minimum version of Blender that should show this asset."""

    until: str | None = None
    """First version of Blender that should NOT show this asset."""


@dataclass
class CatalogV1:
    """An asset catalog, which can be represented by one or more UUIDs."""

    path: str
    uuids: list[str]
    simple_name: str | None = None


@dataclass
class FileV1:
    """Single file in the asset library.

    Identified by its relative path in that library.
    """

    path: str
    """Relative path of where this file is located in the asset library."""

    size_in_bytes: int
    hash: str
    """Hash of the file.

    This should be in the format "HASHTYPE:HASH-AS-HEX". Currently only
    the "SHA256" hash type is supported.
    """

    blender_version: str
    """Version of Blender used to write this file.

    Only contains the major and minor version, no patch version ("5.2",
    "6.3", etc. but not "5.2.1").
    """

    url: str | None = None
    """URL where the file can be downloaded.

    If the URL is relative, it is to be interpreted as relative to the
    library's root URL. If the URL is not given, or an empty string, it
    is assumed to be the same as 'path'.
    """


@dataclass
class AssetLibraryMeta:
    """Meta-data of this asset library."""

    api_versions: dict[str, URLWithHash]
    """API versions of this asset library.

    This is reflected in the URLs of all OpenAPI operations except the
    one to get this metadata. A single asset library can expose multiple
    versions, in order to be backward-compatible with older versions of
    Blender. Keys should be "v1", "v2", etc. and their values should be
    a URLWithHash that points to each version's index file.
    """

    name: str
    """Name of this asset library."""

    contact: Contact


@dataclass
class AssetLibraryIndexV1:
    """The available assets at this library."""

    schema_version: str
    """Version number of the used schema.

    This should be the same as the version of this OpenAPI definition,
    as described in its 'info.version' field.
    """

    asset_size_bytes: int
    asset_count: int
    """Total number of assets in this index.

    This is the sum of all `asset_count` fields of each page.
    """

    file_count: int
    """Total number of files in this index.

    This is the sum of all `file_count` fields of each page (after
    deduplication).
    """

    pages: list[URLWithHash]
    """URLs of the individual asset index pages.

    When relative, these are taken as relative to the main server URL
    (i.e. the root of all paths defined in this OpenAPI spec).
    """

    catalogs: list[CatalogV1] | None = None


@dataclass
class AssetLibraryIndexPageV1:
    """Any number of assets."""

    asset_count: int
    """Number of assets in this page.

    This is declared separately, so that a partial JSON parser has this
    information before the entire file is downloaded and parsed.
    """

    file_count: int
    """Number of files in this page.

    This is declared separately, so that a partial JSON parser has this
    information before the entire file is downloaded and parsed.
    """

    assets: list[AssetV1]
    files: list[FileV1]
    """The files that are referenced by the above assets.

    Note that there may be duplication of this information between asset
    pages, as each file can contain multiple assets, and those assets
    might be scattered across multiple pages.
    """


@dataclass
class AssetV1:
    """Representation of a single asset.

    Assets are always Blender data-blocks in some blend file. This asset
    may be stored in the same blend file as other assets, and so it does
    _not_ represent a single downloadable item.
    """

    name: str
    """Name of the Blender data-block."""

    id_type: AssetIDTypeV1
    files: list[str]
    """Relative paths of the files that contain this asset.

    The first entry in the list MUST contain the asset data-block
    itself, while the remaining entries can be in any order. These
    relative paths are used to look up more file information in the
    asset library's list of files.
    """

    bl_versions: AssetBlenderVersionsV1
    thumbnail: URLWithHash | None = None
    meta: AssetMetadataV1 | None = None


@dataclass
class AssetMetadataV1:
    """Metadata of an asset, as defined by Blender's `AssetMeta` DNA struct.

    Fields should either be non-empty or absent.
    """

    catalog_id: str | None = None
    """The catalog UUID that contains this asset.

    Having the UUID here makes it easier to create a per-blendfile
    .cats.txt file, if that's ever necessary.
    """

    preferred_import_method: str | None = None
    """The import method preferred by this asset.

    For example, base meshes for sculpting can declare they should
    always be appended, making them instantly usable for sculpting.
    Supports values APPEND, APPEND_REUSE, and ASSET_IMPORT_PACK. These
    are not modeled here as an enum, to aid in forward compatibility of
    this Blender version with future import methods (it'll just ignore
    unsupported methods, instead of rejecting the file as invalid).
    """

    tags: list[str] | None = None
    author: str | None = None
    description: str | None = None
    license: str | None = None
    copyright: str | None = None
    properties: CustomPropertiesV1 | None = None


type CustomPropertiesV1 = list[CustomPropertyV1]
"""Arbitrary custom properties of the asset.

Keys are the property names.
"""


@dataclass
class CustomPropertyV1:
    """Single 'custom property' value of the asset.

    The value should be compatible with the given type; GROUP properties
    should be represented as `CustomPropertiesV1` object again. Arrays
    should specify an `itemtype`.
    """
    name: str
    type: CustomPropertyTypeV1
    value: CustomPropertiesV1 | list[Any] | float | int | str | bool
    itemtype: CustomPropertyTypeV1 | None = None



# SPDX-FileCopyrightText: 2018-2021 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path


def dll_path(lib_name, lib_display_name) -> Path | None:
    """
    Get the library path, that should be at addon root
    :return: library path.
    """
    import bpy

    if sys.platform == 'win32':
        library_name = '{}.dll'.format(lib_name)
        # Bundled beside the add-on's `__init__.py`.
        base = os.path.dirname(sys.modules['io_scene_gltf2'].__file__)
    elif sys.platform == 'darwin':
        library_name = 'lib{}.dylib'.format(lib_name)
        # Bundled beside the add-on's `__init__.py`.
        base = os.path.dirname(sys.modules['io_scene_gltf2'].__file__)
    else:
        # Linux, BSD & other UNIX-like systems.
        library_name = 'lib{}.so'.format(lib_name)
        if (system_libs := bpy.utils.resource_path('SYSTEM_LIBS')):
            # System installs place the library beside the add-on, under the system-libs path.
            base = os.path.join(system_libs, 'scripts', 'addons_core', 'io_scene_gltf2')
        elif local := bpy.utils.resource_path('LOCAL'):
            # Portable builds bundle the library in `lib`, beside the shared libraries it depends on.
            base = os.path.join(os.path.dirname(local), 'lib')
        else:
            return None

    return Path(os.path.join(base, library_name))


def dll_exists(lib_name, lib_display_name) -> bool:
    """
    Checks whether the DLL path exists.
    :return: True if the DLL exists.
    """
    path = dll_path(lib_name, lib_display_name)
    if path is None:
        print('ERROR', '{} is not available because its location could not be determined'.format(lib_display_name))
        return False
    exists = path.exists() and path.is_file()
    if exists:
        print('INFO', '{} is available, use library at %s'.format(lib_display_name) % path.absolute())
    else:
        print(
            'ERROR',
            '{} is not available because library could not be found at %s'.format(lib_display_name) %
            path.absolute())
    return exists



import bpy
strip = bpy.context.active_strip

strip.wrap_width = 1.0
strip.use_bold = False
strip.use_italic = False
strip.font_size = 100.0
strip.color = (1.0, 1.0, 1.0, 1.0)
strip.use_outline = True
strip.outline_color = (0.0, 0.0, 0.0, 0.7)
strip.outline_width = 0.1
strip.use_shadow = True
strip.shadow_color = (0.0, 0.0, 0.0, 0.7)
strip.shadow_angle = 1.134
strip.shadow_offset = 0.04
strip.shadow_blur = 0.1
strip.use_box = False
strip.location = (0.5, 0.5)
strip.alignment_x = 'CENTER'
strip.anchor_x = 'CENTER'
strip.anchor_y = 'CENTER'



# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Wrapper around cattrs."""

__all__ = [
    "ValidatingParser",
    "APIModel",
]

import dataclasses
import json
from typing import Any, Type, TypeVar

import cattrs
import cattrs.preconf.json

from . import blender_asset_library_openapi as api_models

# There is no common base class for dataclasses, so this type variable will have to act as a stand-in.
APIModel = TypeVar("APIModel")


class ValidatingParser:
    """Wrapper around cattrs, caching the cattrs converter."""

    _converter: cattrs.preconf.json.JsonConverter

    def __init__(self) -> None:
        self._converter = cattrs.preconf.json.JsonConverter(omit_if_default=True)

        # Register a custom unstructure hook for the type of `CustomPropertyV1.value`.
        #
        # NOTE: this MUST register the 'final' type, and cannot use
        # `CustomProperties` as an alias for `dict[str, CustomProperty]`. It
        # won't be found. It also has to include None in the union for some
        # reason, even though that's not declared in `CustomPropertyV1.value`.
        #
        # Basically cattrs told me to register a structure hook for this
        # specific type, and so that's what I (Sybren) did.
        self._converter.register_structure_hook(
            api_models.CustomPropertiesV1 | list[Any] | float | int | str | bool,
            lambda value, _: value,
        )

    def parse_and_validate(self, model_class: Type[APIModel], json_payload: bytes | str) -> APIModel:
        """Parse & validate the JSON data, returning an instance of the given model class.

        :raises json.JSONDecodeError: if the payload is not formatted as JSON.
        :raises cattrs.errors.ClassValidationError: if the payload doesn't pass
            validation and can't be converted to the given model class.
        """

        json_doc = json.loads(json_payload)
        return self._converter.structure(json_doc, model_class)

    def dumps(self, model_instance: Any) -> str:
        """Convert the model instance to JSON, returning it as string."""
        assert dataclasses.is_dataclass(model_instance), f"{model_instance} is not a dataclass"
        return self._converter.dumps(model_instance, indent=2)



# SPDX-FileCopyrightText: 2018-2026 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0


from .....io.com.gltf2_io_extensions import Extension
from ..search_node_tree import get_socket_from_gltf_material_node, get_const_from_default_value_socket
import bpy


def export_dispersion(bmat, extensions, export_settings):
    export_settings['current_texture_transform'] = {}
    # If no volume extension, no dispersion extension export
    if "KHR_materials_volume" not in extensions:
        return None

    dispersion_socket = get_socket_from_gltf_material_node(
        bmat.get_used_material().node_tree, 'Dispersion')
    if dispersion_socket.socket is None:
        # If no dispersion (here because there is no glTF Material Output node), no dispersion extension export
        return None

    dispersion_extension = {}
    if isinstance(dispersion_socket.socket, bpy.types.NodeSocket):
        dispersion, path = get_const_from_default_value_socket(dispersion_socket, kind='VALUE')
        if dispersion != 0.0:
            dispersion_extension['dispersion'] = dispersion

        # Storing path for KHR_animation_pointer
        if path is not None:
            path_ = {}
            path_['length'] = 1
            path_['path'] = "/materials/XXX/extensions/KHR_materials_dispersion/dispersion"
            export_settings['current_paths'][path] = path_

    return Extension('KHR_materials_dispersion', dispersion_extension, False)



# SPDX-FileCopyrightText: 2026 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

import numpy as np


def extract_attribute_data(attribute, size, type_, data_type, domain, length, export_settings):

    data = None

    if domain == 'CORNER':
        data = np.empty(size * length, dtype=type_)
    elif domain == 'POINT':
        data = np.empty(size * length, dtype=type_)
    elif domain == "EDGE":
        data = np.empty(size * length, dtype=type_)
    elif domain == 'FACE':
        data = np.empty(size * length, dtype=type_)
    else:
        export_settings['log'].error("domain not known")

    if data_type == "BYTE_COLOR":
        attribute.data.foreach_get('color', data)
        data = data.reshape(-1, length)
    elif data_type == "INT8":
        attribute.data.foreach_get('value', data)
        data = data.reshape(-1, length)
    elif data_type == "FLOAT2":
        attribute.data.foreach_get('vector', data)
        data = data.reshape(-1, length)
    elif data_type == "BOOLEAN":
        attribute.data.foreach_get('value', data)
        data = data.reshape(-1, length)
    elif data_type == "STRING":
        attribute.data.foreach_get('value', data)
        data = data.reshape(-1, length)
    elif data_type == "FLOAT_COLOR":
        attribute.data.foreach_get('color', data)
        data = data.reshape(-1, length)
    elif data_type == "FLOAT_VECTOR":
        attribute.data.foreach_get('vector', data)
        data = data.reshape(-1, length)
    elif data_type == "QUATERNION":
        attribute.data.foreach_get('value', data)
        data = data.reshape(-1, length)
    elif data_type == "FLOAT4X4":
        attribute.data.foreach_get('value', data)
        data = data.reshape(-1, length)
    elif data_type == "INT":
        attribute.data.foreach_get('value', data)
        data = data.reshape(-1, length)
    elif data_type == "FLOAT":
        attribute.data.foreach_get('value', data)
        data = data.reshape(-1, length)
    else:
        export_settings['log'].error("blender type not found " + data_type)

    return data



# SPDX-FileCopyrightText: 2018-2021 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

import base64

from ...io.com import gltf2_io
from ...io.exp import binary_data as gltf2_io_binary_data


class Buffer:
    """Class representing binary data for use in a glTF file as 'buffer' property."""

    def __init__(self, is_glb, meshopt_extension, buffer_index=0, initial_data=None):
        self.is_glb = is_glb
        self.meshopt_extension = meshopt_extension
        self.__data = bytearray(b"")
        if initial_data is not None:
            self.__data = bytearray(initial_data.tobytes())
        self.__buffer_index = buffer_index
        self.__fake_bytelength = 0

    def add_fake_bytelength(self, byte_length):
        """used for meshopt compression fallback"""
        self.__fake_bytelength += byte_length

    def get_fake_bytelength(self):
        """used for meshopt compression fallback"""
        return self.__fake_bytelength

    def add_and_get_view(self, binary_data: gltf2_io_binary_data.BinaryData,
                         additional_buffer=None) -> gltf2_io.BufferView:
        """Add binary data to the buffer. Return a glTF BufferView."""

        # If there is an additional buffer => We are exporting with meshopt
        # It means that compressed data will go in the additional buffer
        # And uncomppressed data will also go in the additional buffer,
        # as only fallback data must go in the main buffer (bufferView definition, but without any real data.

        # if hasattr(binary_data, 'extensions') => Compressed data,
        # So populate main bufferview with main buffer, extension buffer view with additional buffer
        # else => no compressed data, so populate main bufferview with additional (compressed) buffer, no extension

        if not additional_buffer:
            offset = len(self.__data)
            self.__data.extend(binary_data.data)

        else:
            offset = len(additional_buffer.__data)
            fake_byte_length = self.__fake_bytelength  # Calculate the offset for the fallback data
            if not hasattr(binary_data, 'extensions'):
                additional_buffer.__data.extend(binary_data.data)
            else:
                self.add_fake_bytelength(binary_data.byte_length)
                # We also need to have a padding in the main buffer for the fallback data,
                # even if it doesn't contain real data, to have the correct offset for the
                # fallback bufferview
                padding = (4 - (binary_data.byte_length % 4)) % 4
                self.add_fake_bytelength(padding)

        length = binary_data.byte_length

        # offsets should be a multiple of 4 --> therefore add padding if necessary
        padding = (4 - (length % 4)) % 4
        if not additional_buffer:
            self.__data.extend(b"\x00" * padding)
        else:
            if not hasattr(binary_data, 'extensions'):
                additional_buffer.__data.extend(b"\x00" * padding)

        if not additional_buffer:
            buffer_index = self.__buffer_index
        else:
            if not self.is_glb:
                if hasattr(binary_data, 'extensions'):
                    buffer_index = self.__buffer_index
                else:
                    buffer_index = additional_buffer.__buffer_index
            else:
                # For glb, make sure the compressed buffer is first, and the fallback second
                if hasattr(binary_data, 'extensions'):
                    buffer_index = 1  # fallback buffer
                else:
                    buffer_index = 0  # compressed / data buffer

        buffer_view = gltf2_io.BufferView(
            buffer=buffer_index,
            byte_length=length,
            byte_offset=fake_byte_length if hasattr(binary_data, 'extensions') else offset,
            byte_stride=None,
            extensions=None,
            extras=None,
            name=None,
            target=binary_data.bufferViewTarget
        )

        if additional_buffer is not None and hasattr(binary_data, 'extensions'):
            # KHR/EXT_meshopt_compression
            compressed_binary_data = binary_data.extensions[self.meshopt_extension]['buffer']
            additional_buffer.__data.extend(compressed_binary_data)
            length = len(compressed_binary_data)
            padding = (4 - (length % 4)) % 4
            additional_buffer.__data.extend(b"\x00" * padding)

            buffer_view.extensions = binary_data.extensions

            # For glb, make sure the real data (compressed or not) is in the first
            # buffer, and the fallback in the second buffer
            if not self.is_glb:
                buffer_view.extensions[self.meshopt_extension]['buffer'] = additional_buffer.__buffer_index
            else:
                buffer_view.extensions[self.meshopt_extension]['buffer'] = 0

            buffer_view.extensions[self.meshopt_extension]['byteOffset'] = offset

        return buffer_view

    @property
    def byte_length(self):
        return len(self.__data)

    def to_bytes(self):
        return self.__data

    def clear(self):
        self.__data = b""

    def to_embed_string(self):
        return 'data:application/octet-stream;base64,' + base64.b64encode(self.__data).decode('ascii')



# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
import os
import re
import shutil
import unicodedata
import urllib.parse
from pathlib import Path

import bpy

from . import blender_asset_library_openapi as api_models
from . import hashing

log = logging.getLogger(__name__)


def list_assets(blendfile: Path, asset_library_root: Path) -> tuple[api_models.FileV1, list[api_models.AssetV1]]:
    # Start by erasing everything from memory.
    bpy.ops.wm.read_homefile(use_factory_startup=True, use_empty=True, load_ui=False)

    blendfile_info = _blendfile_info(blendfile, asset_library_root)

    # Tell Blender to only load asset data-blocks.
    with bpy.data.libraries.load(str(blendfile), assets_only=True) as (
        data_from,
        data_to,
    ):
        for attr in dir(data_to):
            setattr(data_to, attr, getattr(data_from, attr))

        # Convert the Blender version to a string.
        blend_version = ".".join(map(str, data_from.version))
        blendfile_info.blender_version = blend_version

    # Get the last modification timestamp of the blend file, to compare against
    # the thumbnails.
    thumbnail_dir = blendfile.with_name(blendfile.stem + "_thumbnails")
    blend_stat = blendfile.stat()

    thumbnail_timestamper = thumbnail_dir / ".last_modified"
    if thumbnail_timestamper.exists():
        thumb_mtime = thumbnail_timestamper.stat().st_mtime
        should_write_thumbnails = abs(blend_stat.st_mtime - thumb_mtime) > 0.001
    else:
        should_write_thumbnails = True

    if should_write_thumbnails:
        # Remove the entire thumbnail tree, so that thumbnails of deleted assets
        # are also deleted. All thumbnails are going to be re-written anyway.
        log.debug("thumbnails will be exported to %s", thumbnail_dir)
        assert thumbnail_dir
        if Path(thumbnail_dir.root) == thumbnail_dir:
            raise RuntimeError(f"Refusing to remove a root directory: {thumbnail_dir}")
        if thumbnail_dir.exists():
            shutil.rmtree(thumbnail_dir)

    # Collect the asset data.
    assets: list[api_models.AssetV1] = []
    for attr in dir(data_to):
        if attr == 'version':
            continue
        datablocks = getattr(data_from, attr)
        datablocks_assets = _find_assets(
            asset_library_root,
            blendfile_info,
            datablocks,
            thumbnail_dir,
            should_write_thumbnails,
        )
        assets.extend(datablocks_assets)

    # After processing is done, set the thumbnail dir mtime to that of the
    # blendfile. By tracking the mtime of the directory itself, not every
    # individual thumbnail needs to be time-checked.
    thumbnail_dir.mkdir(exist_ok=True, parents=True)
    thumbnail_timestamper.touch(exist_ok=True)
    os.utime(thumbnail_timestamper, (blend_stat.st_atime, blend_stat.st_mtime))

    return blendfile_info, assets


def _find_assets(
    asset_library_root: Path,
    file: api_models.FileV1,
    datablocks: bpy.types.BlendData,
    thumbnail_dir: Path,
    should_write_thumbnails: bool,
) -> list[api_models.AssetV1]:

    # TODO: when multiple files are supported, take the maximum of the files.
    bl_versions = api_models.AssetBlenderVersionsV1(
        min='.'.join(file.blender_version.split('.')[:2]),
    )

    assets = []
    for datablock in datablocks:
        asset_data: bpy.types.AssetData = datablock.asset_data
        if not asset_data:
            continue

        thumbnail_path = _thumbnail_path(datablock, thumbnail_dir)
        if thumbnail_path and should_write_thumbnails:
            _save_thumbnail(datablock, thumbnail_path)

        if thumbnail_path and thumbnail_path.exists():
            as_posix = thumbnail_path.relative_to(asset_library_root).as_posix()
            thumbnail = api_models.URLWithHash(
                url=urllib.parse.quote(as_posix),
                hash=hashing.hash_file(thumbnail_path),
            )
        else:
            thumbnail = None

        asset = api_models.AssetV1(
            name=datablock.name,
            id_type=datablock.id_type,
            files=[file.path],
            thumbnail=thumbnail,
            bl_versions=bl_versions,
            meta=_get_asset_meta(asset_data),
        )

        assets.append(asset)
    return assets


def _get_asset_meta(asset_data: bpy.types.AssetData) -> api_models.AssetMetadataV1 | None:
    # Only set the fields that have a value. That way we can detect whether
    # none of them are set, and prevent the empty metadata from being
    # included.
    meta = api_models.AssetMetadataV1()
    if asset_data.catalog_id and asset_data.catalog_id != "00000000-0000-0000-0000-000000000000":
        meta.catalog_id = asset_data.catalog_id
    if asset_data.tags:
        meta.tags = [tag.name for tag in asset_data.tags]
    if asset_data.author:
        meta.author = asset_data.author
    if asset_data.description:
        meta.description = asset_data.description
    if asset_data.license:
        meta.license = asset_data.license
    if asset_data.copyright:
        meta.copyright = asset_data.copyright
    if asset_data.use_preferred_import_method:
        meta.preferred_import_method = asset_data.preferred_import_method

    # Convert custom properties.
    import rna_prop_ui

    custom_props: api_models.CustomPropertiesV1 = []
    for prop_name, prop_value in asset_data.items():
        is_array = isinstance(prop_value, rna_prop_ui.ARRAY_TYPES) and len(prop_value) > 0
        item_value = prop_value[0] if is_array else prop_value

        match item_value:
            case bool():
                value_type = api_models.CustomPropertyTypeV1.IDP_BOOL
            case int():
                value_type = api_models.CustomPropertyTypeV1.IDP_INT
            case str():
                value_type = api_models.CustomPropertyTypeV1.IDP_STRING
            case float():
                value_type = api_models.CustomPropertyTypeV1.IDP_FLOAT
            case _:
                # Unsupported type, just ignore it.
                continue

        if is_array:
            custom_prop = api_models.CustomPropertyV1(
                name=prop_name,
                type=api_models.CustomPropertyTypeV1.IDP_ARRAY,
                value=list(prop_value),
                itemtype=value_type,
            )
        else:
            custom_prop = api_models.CustomPropertyV1(
                name=prop_name, type=value_type, value=prop_value
            )

        custom_props.append(custom_prop)

    if custom_props:
        meta.properties = custom_props

    if meta == api_models.AssetMetadataV1():
        return None
    return meta


def _save_thumbnail(datablock: bpy.types.ID, thumbnail_path: Path) -> None:
    """Save the internal preview thumbnail as a WebP image."""

    # Get the preview image size.
    width: int = datablock.preview.image_size[0]
    height: int = datablock.preview.image_size[1]

    if not (width > 0 and height > 0):
        return

    thumbnail_path.parent.mkdir(exist_ok=True, parents=True)

    log.debug("Writing thumbnail: %s", thumbnail_path)
    try:
        # Create a new image in Blender to store the preview.
        image: bpy.types.Image = bpy.data.images.new(
            thumbnail_path.stem, width, height, alpha=True
        )

        # Assign the pixel data from the preview to the new image.
        # image.pixels = [p for p in datablock.preview.image_pixels_float]
        image.pixels[:] = datablock.preview.image_pixels_float

        # Save the image to disk.
        image.file_format = "WEBP"
        image.save(filepath=str(thumbnail_path), quality=80)

        # Remove the image from Blender data after saving to free memory.
        bpy.data.images.remove(image)
    except Exception as e:
        print(f"Failed to save thumbnail for {datablock.name}: {e}")


def _thumbnail_path(datablock: bpy.types.ID, thumbnail_dir: Path) -> Path | None:
    """Return the path for this datablock's thumbnail, or None if it has none."""

    if not datablock.preview:
        return None

    datablock_safe = _name_to_filename(datablock.name)
    thumbnail_path: Path = (
        thumbnail_dir / datablock.id_type.title() / f"{datablock_safe}.webp"
    )

    return thumbnail_path


_re_safe_filename_nonword = re.compile(r'[^\w\s_-]')
_re_safe_filename_dashspace = re.compile(r'[-\s]+')


def _name_to_filename(value: str) -> str:
    """Convert a string into something that should be safe as filename."""

    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = _re_safe_filename_nonword.sub('', value.lower())
    return _re_safe_filename_dashspace.sub('-', value).strip('-_')


def _blendfile_info(filepath: Path, asset_library_root: Path) -> api_models.FileV1:
    stat = filepath.stat()

    relative_posix = filepath.relative_to(asset_library_root).as_posix()
    file_url: str | None = urllib.parse.quote(relative_posix)

    if file_url == relative_posix:
        # Optimization: if the file path is URL-safe, it can be used as the URL
        # and there is no need to include this URL explicitly.
        file_url = None

    return api_models.FileV1(
        path=relative_posix,
        url=file_url,
        hash=hashing.hash_file(filepath),
        size_in_bytes=stat.st_size,
        blender_version="",  # Determined later when the file is opened to find assets.
    )



# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations


# The function below is (un)registered from scripts/addons_core/bl_pkg/__init__.py:
def asset_listing_main(args: list[str]) -> int:
    """Run the `blender -c asset_listing` CLI command.

    This is late-importing the cli module, so that it (and its
    dependencies) are only imported when actually used.
    """
    import traceback
    from . import cli

    try:
        cli.main(args)
    except SystemExit as ex:
        if isinstance(ex.code, int):
            return ex.code
        return 2
    except BaseException:
        traceback.print_exc()
        return 1
    return 0



# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Parser for Blender's asset catalog files.

It would be better if there was an RNA API for this, but for now this is faster
to implement.
"""

from __future__ import annotations

import dataclasses
import uuid
from pathlib import Path, PurePosixPath

from . import blender_asset_library_openapi as api_models

SUPPORTED_VERSION = 1


@dataclasses.dataclass(frozen=True)
class AssetCatalog:
    uuid: str
    path: PurePosixPath
    simple_name: str


def parse_catalogs(library_path: Path) -> list[api_models.CatalogV1]:
    """Parse all asset catalog files in the asset library.

    Returns a collection of all asset catalogs in the library, as a mapping from
    UUID to the catalog.

    If there are multiple catalog definition files, they will be merged
    together.
    """
    # First use a mapping from UUID to the AssetCatalog, to ensure that each
    # UUID only maps to a single path.
    catalogs_by_uuid: dict[str, AssetCatalog] = {}

    for file in library_path.rglob('*.cats.txt'):
        file_cats = _parse_catalog(file)
        catalogs_by_uuid.update(file_cats)

    # Group catalogs by their path, to make the returned list compatible with
    # the API model.
    asset_cats_by_path: dict[PurePosixPath, api_models.CatalogV1] = {}
    for cat in catalogs_by_uuid.values():
        try:
            api_catalog = asset_cats_by_path[cat.path]
        except KeyError:
            asset_cats_by_path[cat.path] = api_models.CatalogV1(
                path=cat.path.as_posix(),
                uuids=[cat.uuid],
                simple_name=cat.simple_name,
            )
        else:
            api_catalog.uuids.append(cat.uuid)

    return sorted(asset_cats_by_path.values(), key=lambda api_cat: api_cat.path)


def _parse_catalog(catalog_filepath: Path) -> dict[str, AssetCatalog]:
    # Mapping from UUID to the AssetCatalog.
    catalogs: dict[str, AssetCatalog] = {}

    with catalog_filepath.open('r', encoding='utf-8') as infile:
        for line in infile:
            line = line.strip()
            if not line or line.startswith('#'):
                continue

            # Check the declared version, and simply ignore the file if it is
            # not supported.
            if line.startswith('VERSION '):
                _, version_as_str = line.split(maxsplit=1)
                if version_as_str != str(SUPPORTED_VERSION):
                    msg = "{}: this version of Blender does not support catalog file version {!r}"
                    print(msg.format(catalog_filepath, version_as_str))
                    return {}
                continue

            parts = line.split(':', maxsplit=2)
            if len(parts) < 2:
                # It's ok for the 'simple name' part to be missing, but if more is missing, this is not a valid file.
                msg = "{}: this does not seem to be an asset catalog file, ignoring it (line {!r} is not as expected)"
                print(msg.format(catalog_filepath, line))
                return {}

            cat = AssetCatalog(
                uuid=parts[0],
                path=PurePosixPath(parts[1]),
                simple_name=parts[2] if len(parts) >= 3 else "",
            )
            catalogs[cat.uuid] = cat

    return catalogs


_ASSET_CATS_HEADER = """# This is an Asset Catalog Definition file for Blender.
#
# Empty lines and lines starting with `#` will be ignored.
# The first non-ignored line should be the version indicator.
# Other lines are of the format "UUID:catalog/path/for/assets:simple catalog name"
#
# Remote Asset Library: {library_name!s}

VERSION 1
"""


def write(catalogs: list[api_models.CatalogV1], catalog_filepath: Path,
          asset_library_meta: api_models.AssetLibraryMeta) -> None:
    """Create a catalog file from the list of catalogs."""
    import re

    # TODO: this really should be using an RNA API.

    # Sanitize the library name, as it should not contain any newlines for the Asset Catalog Definition File to be
    # valid. To be on the safe side, just collapse all white-space to spaces. Same for colons, those are used as field
    # separators and shouldn't be included in any of the fields themselves.
    unwanted_chars_re = re.compile(r'[\s:]+')
    lib_name = unwanted_chars_re.sub(' ', asset_library_meta.name)

    header = _ASSET_CATS_HEADER.format(library_name=lib_name)

    with catalog_filepath.open("w", encoding="utf8") as catfile:
        print(header, file=catfile)

        for cat in sorted(catalogs, key=lambda cat: cat.path):
            for cat_uuid_str in cat.uuids:
                # Sanitize the catalogs before writing them.
                try:
                    cat_uuid = uuid.UUID(cat_uuid_str)
                except ValueError:
                    print("Asset Library has invalid UUID ({uuid!r}) for catalog {path!r}, skipping".format(
                        uuid=cat_uuid_str, path=cat.path))
                    continue

                cat_path = unwanted_chars_re.sub(' ', cat.path)

                if isinstance(cat.simple_name, str):
                    cat_simple_name = unwanted_chars_re.sub(' ', cat.simple_name)
                else:
                    cat_simple_name = ""

                print("{!s}:{!s}:{!s}".format(cat_uuid, cat_path, cat_simple_name), file=catfile)



# SPDX-FileCopyrightText: 2026 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
from pathlib import Path

from _bpy_internal.http import downloader as http_dl


class ExtraFileMetadataProvider(http_dl.MetadataProvider):
    """HTTP Metadata provider that can check an extra file.

    This is to support the following file sets:

    - `file.json`: Actual JSON file read by Blender. Is assumed to be validated.
    - `file-unsafe.json`: JSON file as downloaded. Must be validated before use.
    - `file-unsafe.json~`: The above file while it's being downloaded. Not yet
      complete JSON.

    The downloader will get the request to download to `file-unsafe.json`.
    However, if `file.json` is still fresh (i.e. the HTTP metadata for the URL
    is applicable to that file), the downloader should be able to do a
    conditional download (instead of an unconditional one).

    This is implemented as a wrapper for any other MetadataProvider, rather than
    subclassing a specific one, so that it's independent of the underlying
    logic.
    """

    _wrapped: http_dl.MetadataProvider
    _logger: logging.Logger

    def __init__(self, wrapped: http_dl.MetadataProvider) -> None:
        self._wrapped = wrapped
        self._logger = logging.getLogger(__name__ + ".ExtraFileMetadataProvider")

    def save(self, http_req_descr: http_dl.RequestDescription, meta: http_dl.HTTPMetadata) -> None:
        self._wrapped.save(http_req_descr, meta)

    def load(self, http_req_descr: http_dl.RequestDescription) -> http_dl.HTTPMetadata | None:
        return self._wrapped.load(http_req_descr)

    def is_valid(
            self,
            meta: http_dl.HTTPMetadata,
            http_req_descr: http_dl.RequestDescription,
            local_path: Path) -> bool:
        # This assumes that the download is saved to the "unsafe" location, and
        # we have to check the metadata on the "safe" location as well.

        if self._wrapped.is_valid(meta, http_req_descr, local_path):
            self._logger.info("HTTP metadata is valid for %s", local_path)
            return True

        safe_filename = unsafe_to_safe_filename(local_path)
        if safe_filename == local_path:
            # There is no different filename to check, so let's stick to the
            # result of the first is_valid() call.
            self._logger.info("HTTP metadata is invalid for %s", local_path)
            return False

        if self._wrapped.is_valid(meta, http_req_descr, safe_filename):
            self._logger.info("HTTP metadata is valid for %s", safe_filename)
            return True

        self._logger.info("HTTP metadata is valid for neither %s nor %s", local_path, safe_filename)
        return False

    def forget(self, http_req_descr: http_dl.RequestDescription) -> None:
        self._wrapped.forget(http_req_descr)


def unsafe_to_safe_filename(unsafe_file_path: Path) -> Path:
    """path/to/some_file.unsafe-json -> path/to/some_file.json"""
    # The suffix is changed, and not the stem, so that globs like "*.json" do not see the unsafe files.
    return unsafe_file_path.with_suffix(unsafe_file_path.suffix.replace('unsafe-', ''))


def safe_to_unsafe_filename(safe_file_path: Path | str) -> Path:
    """path/to/some_file.json -> path/to/some_file.unsafe-json"""
    if isinstance(safe_file_path, str):
        safe_file_path = Path(safe_file_path)

    # path.suffix includes the leading period, so it's something like ".json".
    return safe_file_path.with_suffix('.unsafe-' + safe_file_path.suffix[1:])
